import click


# TODO: type the return value
def try_import_gcp_secretmanager():
    try:
        from google.cloud import secretmanager

        return secretmanager
    except ImportError:
        raise click.ClickException(
            "pip package `google-cloud-secret-manager` is not installed locally on this machine but required "
            "for the command. Please install with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_discovery():
    try:
        from googleapiclient import discovery

        return discovery
    except ImportError:
        raise click.ClickException(
            "pip package `google-api-python-client` is not installed locally on this machine but required "
            "for the command. Please install with `pip install 'anyscale[gcp]'`."
        )


def try_import_gcp_utils():
    try:
        from anyscale.utils.gcp_utils import (
            get_application_default_credentials,
            get_gcp_filestore_config,
        )

        return get_application_default_credentials, get_gcp_filestore_config
    except ImportError:
        raise click.ClickException(
            "Please install GCP-related packages with `pip install anyscale[gcp]`."
        )


def try_import_gcp_client_factory():
    try:
        from anyscale.gcp_verification import GoogleCloudClientFactory

        return GoogleCloudClientFactory
    except ImportError:
        raise click.ClickException(
            "Please install GCP-related packages with `pip install anyscale[gcp]`."
        )


def try_import_gcp_logger():
    try:
        from anyscale.gcp_verification import GCPLogger

        return GCPLogger
    except ImportError:
        raise click.ClickException(
            "Please install GCP-related packages with `pip install anyscale[gcp]`."
        )


def try_import_gcp_verify_lib():
    try:
        import anyscale.gcp_verification as verify_lib

        return verify_lib
    except ImportError:
        raise click.ClickException(
            "Please install GCP-related packages with `pip install anyscale[gcp]`."
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class Cloud(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'provider': 'CloudProviders',
        'region': 'str',
        'credentials': 'str',
        'config': 'CloudConfig',
        'is_k8s': 'bool',
        'is_aioa': 'bool',
        'availability_zones': 'list[str]',
        'is_bring_your_own_resource': 'bool',
        'is_private_cloud': 'bool',
        'cluster_management_stack_version': 'ClusterManagementStackVersions',
        'id': 'str',
        'type': 'CloudTypes',
        'creator_id': 'str',
        'created_at': 'datetime',
        'status': 'CloudStatus',
        'state': 'CloudState',
        'version': 'CloudVersion'
    }

    attribute_map = {
        'name': 'name',
        'provider': 'provider',
        'region': 'region',
        'credentials': 'credentials',
        'config': 'config',
        'is_k8s': 'is_k8s',
        'is_aioa': 'is_aioa',
        'availability_zones': 'availability_zones',
        'is_bring_your_own_resource': 'is_bring_your_own_resource',
        'is_private_cloud': 'is_private_cloud',
        'cluster_management_stack_version': 'cluster_management_stack_version',
        'id': 'id',
        'type': 'type',
        'creator_id': 'creator_id',
        'created_at': 'created_at',
        'status': 'status',
        'state': 'state',
        'version': 'version'
    }

    def __init__(self, name=None, provider=None, region=None, credentials=None, config=None, is_k8s=False, is_aioa=False, availability_zones=None, is_bring_your_own_resource=None, is_private_cloud=False, cluster_management_stack_version=None, id=None, type=None, creator_id=None, created_at=None, status=None, state=None, version=None, local_vars_configuration=None):  # noqa: E501
        """Cloud - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._provider = None
        self._region = None
        self._credentials = None
        self._config = None
        self._is_k8s = None
        self._is_aioa = None
        self._availability_zones = None
        self._is_bring_your_own_resource = None
        self._is_private_cloud = None
        self._cluster_management_stack_version = None
        self._id = None
        self._type = None
        self._creator_id = None
        self._created_at = None
        self._status = None
        self._state = None
        self._version = None
        self.discriminator = None

        self.name = name
        self.provider = provider
        self.region = region
        self.credentials = credentials
        if config is not None:
            self.config = config
        if is_k8s is not None:
            self.is_k8s = is_k8s
        if is_aioa is not None:
            self.is_aioa = is_aioa
        if availability_zones is not None:
            self.availability_zones = availability_zones
        if is_bring_your_own_resource is not None:
            self.is_bring_your_own_resource = is_bring_your_own_resource
        if is_private_cloud is not None:
            self.is_private_cloud = is_private_cloud
        if cluster_management_stack_version is not None:
            self.cluster_management_stack_version = cluster_management_stack_version
        self.id = id
        self.type = type
        self.creator_id = creator_id
        self.created_at = created_at
        if status is not None:
            self.status = status
        if state is not None:
            self.state = state
        if version is not None:
            self.version = version

    @property
    def name(self):
        """Gets the name of this Cloud.  # noqa: E501

        Name of this Cloud.  # noqa: E501

        :return: The name of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Cloud.

        Name of this Cloud.  # noqa: E501

        :param name: The name of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def provider(self):
        """Gets the provider of this Cloud.  # noqa: E501

        Provider of this Cloud (e.g. AWS).  # noqa: E501

        :return: The provider of this Cloud.  # noqa: E501
        :rtype: CloudProviders
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this Cloud.

        Provider of this Cloud (e.g. AWS).  # noqa: E501

        :param provider: The provider of this Cloud.  # noqa: E501
        :type: CloudProviders
        """
        if self.local_vars_configuration.client_side_validation and provider is None:  # noqa: E501
            raise ValueError("Invalid value for `provider`, must not be `None`")  # noqa: E501

        self._provider = provider

    @property
    def region(self):
        """Gets the region of this Cloud.  # noqa: E501

        Region this Cloud is operating in. This value needs to be supported by this Cloud's provider. (e.g. us-west-2)  # noqa: E501

        :return: The region of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region):
        """Sets the region of this Cloud.

        Region this Cloud is operating in. This value needs to be supported by this Cloud's provider. (e.g. us-west-2)  # noqa: E501

        :param region: The region of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and region is None:  # noqa: E501
            raise ValueError("Invalid value for `region`, must not be `None`")  # noqa: E501

        self._region = region

    @property
    def credentials(self):
        """Gets the credentials of this Cloud.  # noqa: E501

        Credentials needed to interact with this Cloud.  # noqa: E501

        :return: The credentials of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._credentials

    @credentials.setter
    def credentials(self, credentials):
        """Sets the credentials of this Cloud.

        Credentials needed to interact with this Cloud.  # noqa: E501

        :param credentials: The credentials of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and credentials is None:  # noqa: E501
            raise ValueError("Invalid value for `credentials`, must not be `None`")  # noqa: E501

        self._credentials = credentials

    @property
    def config(self):
        """Gets the config of this Cloud.  # noqa: E501

        Additional configurable properties of this Cloud.  # noqa: E501

        :return: The config of this Cloud.  # noqa: E501
        :rtype: CloudConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this Cloud.

        Additional configurable properties of this Cloud.  # noqa: E501

        :param config: The config of this Cloud.  # noqa: E501
        :type: CloudConfig
        """

        self._config = config

    @property
    def is_k8s(self):
        """Gets the is_k8s of this Cloud.  # noqa: E501

        Whether this cloud is managed via Kubernetes.  # noqa: E501

        :return: The is_k8s of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_k8s

    @is_k8s.setter
    def is_k8s(self, is_k8s):
        """Sets the is_k8s of this Cloud.

        Whether this cloud is managed via Kubernetes.  # noqa: E501

        :param is_k8s: The is_k8s of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_k8s = is_k8s

    @property
    def is_aioa(self):
        """Gets the is_aioa of this Cloud.  # noqa: E501

        Whether this cloud is an AIOA cloud.  # noqa: E501

        :return: The is_aioa of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_aioa

    @is_aioa.setter
    def is_aioa(self, is_aioa):
        """Sets the is_aioa of this Cloud.

        Whether this cloud is an AIOA cloud.  # noqa: E501

        :param is_aioa: The is_aioa of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_aioa = is_aioa

    @property
    def availability_zones(self):
        """Gets the availability_zones of this Cloud.  # noqa: E501

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :return: The availability_zones of this Cloud.  # noqa: E501
        :rtype: list[str]
        """
        return self._availability_zones

    @availability_zones.setter
    def availability_zones(self, availability_zones):
        """Sets the availability_zones of this Cloud.

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :param availability_zones: The availability_zones of this Cloud.  # noqa: E501
        :type: list[str]
        """

        self._availability_zones = availability_zones

    @property
    def is_bring_your_own_resource(self):
        """Gets the is_bring_your_own_resource of this Cloud.  # noqa: E501

        Whether the resources of this cloud are provided by the customer.  # noqa: E501

        :return: The is_bring_your_own_resource of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_bring_your_own_resource

    @is_bring_your_own_resource.setter
    def is_bring_your_own_resource(self, is_bring_your_own_resource):
        """Sets the is_bring_your_own_resource of this Cloud.

        Whether the resources of this cloud are provided by the customer.  # noqa: E501

        :param is_bring_your_own_resource: The is_bring_your_own_resource of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_bring_your_own_resource = is_bring_your_own_resource

    @property
    def is_private_cloud(self):
        """Gets the is_private_cloud of this Cloud.  # noqa: E501

        Whether this cloud is a private cloud.  # noqa: E501

        :return: The is_private_cloud of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_private_cloud

    @is_private_cloud.setter
    def is_private_cloud(self, is_private_cloud):
        """Sets the is_private_cloud of this Cloud.

        Whether this cloud is a private cloud.  # noqa: E501

        :param is_private_cloud: The is_private_cloud of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_private_cloud = is_private_cloud

    @property
    def cluster_management_stack_version(self):
        """Gets the cluster_management_stack_version of this Cloud.  # noqa: E501

        The cluster management stack version of the cloud.  # noqa: E501

        :return: The cluster_management_stack_version of this Cloud.  # noqa: E501
        :rtype: ClusterManagementStackVersions
        """
        return self._cluster_management_stack_version

    @cluster_management_stack_version.setter
    def cluster_management_stack_version(self, cluster_management_stack_version):
        """Sets the cluster_management_stack_version of this Cloud.

        The cluster management stack version of the cloud.  # noqa: E501

        :param cluster_management_stack_version: The cluster_management_stack_version of this Cloud.  # noqa: E501
        :type: ClusterManagementStackVersions
        """

        self._cluster_management_stack_version = cluster_management_stack_version

    @property
    def id(self):
        """Gets the id of this Cloud.  # noqa: E501

        Server assigned unique identifier.  # noqa: E501

        :return: The id of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Cloud.

        Server assigned unique identifier.  # noqa: E501

        :param id: The id of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def type(self):
        """Gets the type of this Cloud.  # noqa: E501


        :return: The type of this Cloud.  # noqa: E501
        :rtype: CloudTypes
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this Cloud.


        :param type: The type of this Cloud.  # noqa: E501
        :type: CloudTypes
        """
        if self.local_vars_configuration.client_side_validation and type is None:  # noqa: E501
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501

        self._type = type

    @property
    def creator_id(self):
        """Gets the creator_id of this Cloud.  # noqa: E501

        ID of the User who created this Cloud.  # noqa: E501

        :return: The creator_id of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this Cloud.

        ID of the User who created this Cloud.  # noqa: E501

        :param creator_id: The creator_id of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def created_at(self):
        """Gets the created_at of this Cloud.  # noqa: E501

        Time when this Cloud was created.  # noqa: E501

        :return: The created_at of this Cloud.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Cloud.

        Time when this Cloud was created.  # noqa: E501

        :param created_at: The created_at of this Cloud.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def status(self):
        """Gets the status of this Cloud.  # noqa: E501

        The status of this cloud.  # noqa: E501

        :return: The status of this Cloud.  # noqa: E501
        :rtype: CloudStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this Cloud.

        The status of this cloud.  # noqa: E501

        :param status: The status of this Cloud.  # noqa: E501
        :type: CloudStatus
        """

        self._status = status

    @property
    def state(self):
        """Gets the state of this Cloud.  # noqa: E501

        The state of this cloud.  # noqa: E501

        :return: The state of this Cloud.  # noqa: E501
        :rtype: CloudState
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this Cloud.

        The state of this cloud.  # noqa: E501

        :param state: The state of this Cloud.  # noqa: E501
        :type: CloudState
        """

        self._state = state

    @property
    def version(self):
        """Gets the version of this Cloud.  # noqa: E501

        The version of the cloud.  # noqa: E501

        :return: The version of this Cloud.  # noqa: E501
        :rtype: CloudVersion
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this Cloud.

        The version of the cloud.  # noqa: E501

        :param version: The version of this Cloud.  # noqa: E501
        :type: CloudVersion
        """

        self._version = version

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Cloud):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Cloud):
            return True

        return self.to_dict() != other.to_dict()



from contextlib import contextmanager
from http import HTTPStatus
import json
from typing import Generator, Optional

from anyscale_client.exceptions import ApiException

from anyscale.api_utils.exceptions.job_errors import NoJobRunError


class ExpectedLogRetrievalError(Exception):
    pass


class UnsupportedLogRetrievalMethodError(ExpectedLogRetrievalError):
    pass


class LogRetrievalTimeoutError(ExpectedLogRetrievalError):
    pass


class LogNotFoundOnActiveClusterError(ExpectedLogRetrievalError):
    pass


@contextmanager
def wrap_as_unsupported_log_retrieval_method_error() -> Generator[None, None, None]:
    try:
        yield
    except NoJobRunError as e:
        raise UnsupportedLogRetrievalMethodError from e
    except ApiException as e:
        if e.status == HTTPStatus.BAD_REQUEST:
            raise UnsupportedLogRetrievalMethodError(_get_exception_message(e))
        raise e


def _get_exception_message(e: ApiException) -> Optional[str]:
    return json.loads(e.body).get("error", {}).get("detail")



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ApplyProductionServiceV2APIModel(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'description': 'str',
        'project_id': 'str',
        'version': 'str',
        'canary_percent': 'int',
        'canary_weight': 'int',
        'ray_serve_config': 'object',
        'build_id': 'str',
        'compute_config_id': 'str',
        'rollout_strategy': 'RolloutStrategy',
        'ray_gcs_external_storage_config': 'RayGCSExternalStorageConfig'
    }

    attribute_map = {
        'name': 'name',
        'description': 'description',
        'project_id': 'project_id',
        'version': 'version',
        'canary_percent': 'canary_percent',
        'canary_weight': 'canary_weight',
        'ray_serve_config': 'ray_serve_config',
        'build_id': 'build_id',
        'compute_config_id': 'compute_config_id',
        'rollout_strategy': 'rollout_strategy',
        'ray_gcs_external_storage_config': 'ray_gcs_external_storage_config'
    }

    def __init__(self, name=None, description=None, project_id=None, version=None, canary_percent=None, canary_weight=None, ray_serve_config=None, build_id=None, compute_config_id=None, rollout_strategy=None, ray_gcs_external_storage_config=None, local_vars_configuration=None):  # noqa: E501
        """ApplyProductionServiceV2APIModel - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._description = None
        self._project_id = None
        self._version = None
        self._canary_percent = None
        self._canary_weight = None
        self._ray_serve_config = None
        self._build_id = None
        self._compute_config_id = None
        self._rollout_strategy = None
        self._ray_gcs_external_storage_config = None
        self.discriminator = None

        self.name = name
        if description is not None:
            self.description = description
        self.project_id = project_id
        if version is not None:
            self.version = version
        if canary_percent is not None:
            self.canary_percent = canary_percent
        if canary_weight is not None:
            self.canary_weight = canary_weight
        self.ray_serve_config = ray_serve_config
        self.build_id = build_id
        self.compute_config_id = compute_config_id
        if rollout_strategy is not None:
            self.rollout_strategy = rollout_strategy
        if ray_gcs_external_storage_config is not None:
            self.ray_gcs_external_storage_config = ray_gcs_external_storage_config

    @property
    def name(self):
        """Gets the name of this ApplyProductionServiceV2APIModel.  # noqa: E501

        Name of the Service  # noqa: E501

        :return: The name of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ApplyProductionServiceV2APIModel.

        Name of the Service  # noqa: E501

        :param name: The name of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def description(self):
        """Gets the description of this ApplyProductionServiceV2APIModel.  # noqa: E501

        Description of the Service  # noqa: E501

        :return: The description of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ApplyProductionServiceV2APIModel.

        Description of the Service  # noqa: E501

        :param description: The description of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def project_id(self):
        """Gets the project_id of this ApplyProductionServiceV2APIModel.  # noqa: E501

        Id of the project this Service will start clusters in  # noqa: E501

        :return: The project_id of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this ApplyProductionServiceV2APIModel.

        Id of the project this Service will start clusters in  # noqa: E501

        :param project_id: The project_id of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def version(self):
        """Gets the version of this ApplyProductionServiceV2APIModel.  # noqa: E501

        A version string that represents the version for this service. Will be populated with the hash of the config if not specified.  # noqa: E501

        :return: The version of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this ApplyProductionServiceV2APIModel.

        A version string that represents the version for this service. Will be populated with the hash of the config if not specified.  # noqa: E501

        :param version: The version of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: str
        """

        self._version = version

    @property
    def canary_percent(self):
        """Gets the canary_percent of this ApplyProductionServiceV2APIModel.  # noqa: E501

        A manual target percent for this service. If this field is not set, the service will automatically roll out. If set, this should be a number between 0 and 100. The newly created version will have weight `canary_percent` and the existing version will have `100 - canary_percent`.  # noqa: E501

        :return: The canary_percent of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: int
        """
        return self._canary_percent

    @canary_percent.setter
    def canary_percent(self, canary_percent):
        """Sets the canary_percent of this ApplyProductionServiceV2APIModel.

        A manual target percent for this service. If this field is not set, the service will automatically roll out. If set, this should be a number between 0 and 100. The newly created version will have weight `canary_percent` and the existing version will have `100 - canary_percent`.  # noqa: E501

        :param canary_percent: The canary_percent of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: int
        """

        self._canary_percent = canary_percent

    @property
    def canary_weight(self):
        """Gets the canary_weight of this ApplyProductionServiceV2APIModel.  # noqa: E501

        DEPRECATED: please use 'canary_percent'.  # noqa: E501

        :return: The canary_weight of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: int
        """
        return self._canary_weight

    @canary_weight.setter
    def canary_weight(self, canary_weight):
        """Sets the canary_weight of this ApplyProductionServiceV2APIModel.

        DEPRECATED: please use 'canary_percent'.  # noqa: E501

        :param canary_weight: The canary_weight of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: int
        """

        self._canary_weight = canary_weight

    @property
    def ray_serve_config(self):
        """Gets the ray_serve_config of this ApplyProductionServiceV2APIModel.  # noqa: E501

        The Ray Serve config to use for this service. This config defines your Ray Serve application, and will be passed directly to Ray Serve. You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html  # noqa: E501

        :return: The ray_serve_config of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: object
        """
        return self._ray_serve_config

    @ray_serve_config.setter
    def ray_serve_config(self, ray_serve_config):
        """Sets the ray_serve_config of this ApplyProductionServiceV2APIModel.

        The Ray Serve config to use for this service. This config defines your Ray Serve application, and will be passed directly to Ray Serve. You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html  # noqa: E501

        :param ray_serve_config: The ray_serve_config of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: object
        """
        if self.local_vars_configuration.client_side_validation and ray_serve_config is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_serve_config`, must not be `None`")  # noqa: E501

        self._ray_serve_config = ray_serve_config

    @property
    def build_id(self):
        """Gets the build_id of this ApplyProductionServiceV2APIModel.  # noqa: E501

        The id of the cluster env build. This id will determine the docker image your Service is run using.  # noqa: E501

        :return: The build_id of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: str
        """
        return self._build_id

    @build_id.setter
    def build_id(self, build_id):
        """Sets the build_id of this ApplyProductionServiceV2APIModel.

        The id of the cluster env build. This id will determine the docker image your Service is run using.  # noqa: E501

        :param build_id: The build_id of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and build_id is None:  # noqa: E501
            raise ValueError("Invalid value for `build_id`, must not be `None`")  # noqa: E501

        self._build_id = build_id

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this ApplyProductionServiceV2APIModel.  # noqa: E501

        The id of the compute configuration that you want to use. This id will specify the resources required for your ServiceThe compute template includes a `cloud_id` that must be fixed for each service.  # noqa: E501

        :return: The compute_config_id of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this ApplyProductionServiceV2APIModel.

        The id of the compute configuration that you want to use. This id will specify the resources required for your ServiceThe compute template includes a `cloud_id` that must be fixed for each service.  # noqa: E501

        :param compute_config_id: The compute_config_id of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and compute_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `compute_config_id`, must not be `None`")  # noqa: E501

        self._compute_config_id = compute_config_id

    @property
    def rollout_strategy(self):
        """Gets the rollout_strategy of this ApplyProductionServiceV2APIModel.  # noqa: E501

        Strategy for rollout. The ROLLOUT strategy will deploy your Ray Serve configuration onto a newly started cluster, and then shift traffic over to the new cluster. You can manually control the speed of the rollout using the canary_weight configuration. The IN_PLACE strategy will use Ray Serve in place upgrade to update your existing cluster in place. When using this rollout strategy, you may only change the ray_serve_config field. You cannot partially shift traffic or rollback an in place upgrade. In place upgrades are faster and riskier than rollouts, and we recommend only using them for relatively safe changes (for example, increasing the number of replicas on a Ray Serve deployment). Default strategy is ROLLOUT.  # noqa: E501

        :return: The rollout_strategy of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: RolloutStrategy
        """
        return self._rollout_strategy

    @rollout_strategy.setter
    def rollout_strategy(self, rollout_strategy):
        """Sets the rollout_strategy of this ApplyProductionServiceV2APIModel.

        Strategy for rollout. The ROLLOUT strategy will deploy your Ray Serve configuration onto a newly started cluster, and then shift traffic over to the new cluster. You can manually control the speed of the rollout using the canary_weight configuration. The IN_PLACE strategy will use Ray Serve in place upgrade to update your existing cluster in place. When using this rollout strategy, you may only change the ray_serve_config field. You cannot partially shift traffic or rollback an in place upgrade. In place upgrades are faster and riskier than rollouts, and we recommend only using them for relatively safe changes (for example, increasing the number of replicas on a Ray Serve deployment). Default strategy is ROLLOUT.  # noqa: E501

        :param rollout_strategy: The rollout_strategy of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: RolloutStrategy
        """

        self._rollout_strategy = rollout_strategy

    @property
    def ray_gcs_external_storage_config(self):
        """Gets the ray_gcs_external_storage_config of this ApplyProductionServiceV2APIModel.  # noqa: E501

        Config for the Ray GCS to connect to external storage. If populated, head node fault tolerance will be enabled for this service.  # noqa: E501

        :return: The ray_gcs_external_storage_config of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :rtype: RayGCSExternalStorageConfig
        """
        return self._ray_gcs_external_storage_config

    @ray_gcs_external_storage_config.setter
    def ray_gcs_external_storage_config(self, ray_gcs_external_storage_config):
        """Sets the ray_gcs_external_storage_config of this ApplyProductionServiceV2APIModel.

        Config for the Ray GCS to connect to external storage. If populated, head node fault tolerance will be enabled for this service.  # noqa: E501

        :param ray_gcs_external_storage_config: The ray_gcs_external_storage_config of this ApplyProductionServiceV2APIModel.  # noqa: E501
        :type: RayGCSExternalStorageConfig
        """

        self._ray_gcs_external_storage_config = ray_gcs_external_storage_config

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ApplyProductionServiceV2APIModel):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ApplyProductionServiceV2APIModel):
            return True

        return self.to_dict() != other.to_dict()



import asyncio
import re
from typing import List

import aiohttp

from anyscale.shared_anyscale_utils.utils.asyncio import gather_in_batches


async def _download_logs_concurrently(
    log_chunk_urls: List[str], parallelism: int
) -> str:
    logs_across_chunks = await gather_in_batches(
        parallelism, *[_download_log_from_s3_url(url) for url in log_chunk_urls]
    )
    return "".join(logs_across_chunks)  # type: ignore


async def _download_log_from_ray_json_response(url: str) -> str:
    async with aiohttp.ClientSession() as session:
        response = await asyncio.wait_for(session.get(url), timeout=30)
        logs: str = (await response.json()).get("logs", "")
        return logs


async def _download_log_from_s3_url(url: str) -> str:
    # Note that the URL is presigned, so no token needs to be passed in the request
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.text()


def _remove_ansi_escape_sequences(s: str) -> str:
    # Required as the log may contain ANSI escape sequeneces (e.g. for coloring in the terminal)
    # Regex pattern from https://stackoverflow.com/a/14693789
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", s)



from contextlib import contextmanager
import os
import sys
import time
from typing import Optional

import colorama  # type: ignore
from halo import Halo


_process_start_time = time.time()


class BlockLogger:
    """
    Logger class for the CLI. Supports formatting in blocks if `block_label` is provided to
    the methods. Also supports anyscale.connect style logging if `block_label` is not
    provided.
    """

    def __init__(
        self, log_output: bool = True, t0: float = _process_start_time,
    ) -> None:
        self.t0 = t0
        # Flag to disable all terminal output from CLILogger (useful for SDK)
        self.log_output = log_output
        self.current_block: Optional[str] = None

    def open_block(
        self, block_label: str, block_title: Optional[str] = None, auto_close=False
    ) -> None:
        """
        Prints block title from the given block_label and sets self.current_block. "Output"
        is a generic block that does not need to follow the standard convention.

        if auto_close is set, automatically closes the block before opening a new one
        """
        if not self.log_output:
            return
        assert (
            auto_close or self.current_block is None or self.current_block == "Output"
        ), f"Block {self.current_block} is already open. Please close before opening block {block_label}."

        if auto_close:
            self.close_block()
        self.current_block = block_label
        print(
            f"{colorama.Style.BRIGHT}{colorama.Fore.CYAN}{block_title if block_title else block_label}{colorama.Style.RESET_ALL}",
            file=sys.stderr,
        )

    def close_block(self, block_label: Optional[str] = None) -> None:
        """ Closes the current block
        If a label is specified, it must match the current open block label.
        raise an AssertionError if we try to close a different block

        Prints newline so there is separation before next block is opened.
        """
        if not self.log_output:
            return
        if block_label:
            assert (
                self.current_block == block_label
            ), f"Attempting to close block {block_label}, but block {self.current_block} is currently open."
        self.current_block = None
        print(file=sys.stderr)

    @staticmethod
    def highlight(word: str) -> str:
        return f"{colorama.Style.BRIGHT}{colorama.Fore.MAGENTA}{word}{colorama.Style.RESET_ALL}"

    def zero_time(self) -> None:
        self.t0 = time.time()

    def info(self, *msg: str, block_label: Optional[str] = None) -> None:
        if not self.log_output:
            return
        if block_label:
            # Check block_label if provided.
            assert (
                self.current_block == block_label
            ), f"Attempting to log to block {block_label}, but block {self.current_block} is currently open."
            print(
                *msg, file=sys.stderr,
            )
        else:
            print(
                "{}{}(anyscale +{}){} ".format(
                    colorama.Style.BRIGHT,
                    colorama.Fore.CYAN,
                    self._time_string(),
                    colorama.Style.RESET_ALL,
                ),
                end="",
                file=sys.stderr,
            )
            print(
                *msg, file=sys.stderr,
            )

    def debug(self, *msg: str) -> None:
        if not self.log_output:
            return
        if os.environ.get("ANYSCALE_DEBUG") == "1":
            print(
                "{}{}(anyscale +{}){} ".format(
                    colorama.Style.DIM,
                    colorama.Fore.CYAN,
                    self._time_string(),
                    colorama.Style.RESET_ALL,
                ),
                end="",
            )
            print(*msg)

    def warning(self, *msg: str) -> None:
        if not self.log_output:
            return
        print(
            "{}{}[Warning]{} ".format(
                colorama.Style.NORMAL, colorama.Fore.YELLOW, colorama.Style.RESET_ALL,
            ),
            end="",
            file=sys.stderr,
        )
        print(*msg, file=sys.stderr)

    def _time_string(self) -> str:
        delta = time.time() - self.t0
        hours = 0
        minutes = 0
        while delta > 3600:
            hours += 1
            delta -= 3600
        while delta > 60:
            minutes += 1
            delta -= 60
        output = ""
        if hours:
            output += f"{hours}h"
        if minutes:
            output += f"{minutes}m"
        output += f"{round(delta, 1)}s"
        return output

    def error(self, *msg: str) -> None:
        print(
            "{}{}(anyscale +{}){} ".format(
                colorama.Style.BRIGHT,
                colorama.Fore.RED,
                self._time_string(),
                colorama.Style.RESET_ALL,
            ),
            end="",
            file=sys.stderr,
        )
        print(*msg, file=sys.stderr)


class LogsLogger(BlockLogger):
    """ This logger is used to print customer logs to STDOUT with no decoration
    """

    def log(self, msg: str):
        print(msg)

    def is_interactive_cli_enabled(self) -> bool:
        """Check if shell is interactive
        """

        default_env_var = "1" if sys.stderr.isatty() else "0"
        return os.environ.get("ANYSCALE_CLI_INTERACTIVE_UX", default_env_var) == "1"

    @contextmanager
    def spinner(self, msg: str):
        """ Create a spinner with the starting text.
        To update the text next to the spinner, set `spinner.text = "new_text"`
        The spinner will be stopped when the context is exited.
        Spinner is only enabled for interactive shell
        """

        enable_spinner = self.is_interactive_cli_enabled()

        with Halo(
            text=msg, spinner="dots", stream=sys.stderr, enabled=enable_spinner
        ) as spinner:
            yield spinner



from typing import Optional

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.cloud import get_cloud_id_and_name, get_last_used_cloud
from anyscale.sdk.anyscale_client import (
    ArchiveStatus,
    ComputeTemplateConfig,
    CreateComputeTemplate,
)
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.compute_template import ComputeTemplate
from anyscale.sdk.anyscale_client.models.compute_template_query import (
    ComputeTemplateQuery,
)
from anyscale.utils.cloud_utils import get_organization_default_cloud
from anyscale.utils.name_utils import gen_valid_name


log = BlockLogger()


def get_default_cluster_compute(
    cloud_name: Optional[str],
    project_id: Optional[str],
    api_client: Optional[DefaultApi] = None,
    anyscale_api_client: Optional[DefaultApi] = None,
) -> ComputeTemplate:
    if api_client is None:
        api_client = get_auth_api_client().api_client
    if anyscale_api_client is None:
        anyscale_api_client = get_auth_api_client().anyscale_api_client

    if cloud_name is None:
        default_cloud_name = get_organization_default_cloud(api_client)
        if default_cloud_name:
            cloud_name = default_cloud_name
        else:
            cloud_name = get_last_used_cloud(project_id, anyscale_api_client)

    cloud_id, _ = get_cloud_id_and_name(api_client, cloud_name=cloud_name)
    config_object = anyscale_api_client.get_default_compute_config(cloud_id).result
    compute_template = register_compute_template(config_object, api_client=api_client)
    return compute_template


def get_cluster_compute_from_name(
    cluster_compute_name: str, api_client: Optional[DefaultApi] = None,
) -> ComputeTemplate:
    if api_client is None:
        api_client = get_auth_api_client().api_client
    cluster_computes = api_client.search_compute_templates_api_v2_compute_templates_search_post(
        ComputeTemplateQuery(
            orgwide=True,
            name={"equals": cluster_compute_name},
            include_anonymous=True,
            archive_status=ArchiveStatus.ALL,
        )
    ).results

    if len(cluster_computes) == 0:
        raise ValueError(
            f"The cluster compute template {cluster_compute_name} is not registered."
        )
    return cluster_computes[0]


def register_compute_template(
    config_object: ComputeTemplateConfig, api_client: Optional[DefaultApi] = None,
) -> ComputeTemplate:
    """
    Register compute template with a default name and return the compute template id."""
    if api_client is None:
        api_client = get_auth_api_client().api_client
    created_template = api_client.create_compute_template_api_v2_compute_templates_post(
        create_compute_template=CreateComputeTemplate(
            name=gen_valid_name("cli-autogenerated-config"),
            config=config_object,
            anonymous=True,
        )
    ).result
    return created_template


def get_selected_cloud_id_or_default(
    api_client: Optional[DefaultApi] = None,
    anyscale_api_client: Optional[DefaultApi] = None,
    cluster_compute_id: Optional[str] = None,
    cluster_compute_config: Optional[ClusterComputeConfig] = None,
    cloud_id: Optional[str] = None,
    cloud_name: Optional[str] = None,
):
    """
    Gets cloud_id that is selected for the current command from the
    arguments `cloud_id`, `cloud_name`, `cluster_compute_id`, or `cluster_compute_config`.
    If the cloud_id is not selected through any of these arguments, get the default cloud
    from the default cluster compute.
    """
    api_client = api_client or get_auth_api_client(log_output=False).api_client
    anyscale_api_client = (
        anyscale_api_client or get_auth_api_client(log_output=False).anyscale_api_client
    )
    if cloud_id or cloud_name:
        parent_cloud_id, _ = get_cloud_id_and_name(
            api_client=api_client, cloud_id=cloud_id, cloud_name=cloud_name,
        )
    elif cluster_compute_id:
        parent_cloud_id = anyscale_api_client.get_cluster_compute(
            cluster_compute_id
        ).result.config.cloud_id
    elif cluster_compute_config:
        parent_cloud_id = cluster_compute_config.cloud_id
    else:
        parent_cloud_id = (
            anyscale_api_client.get_default_cluster_compute().result.config.cloud_id
        )
    return parent_cloud_id



from datetime import datetime
import pprint
from typing import Any, Dict, IO, List, Optional, Union

import click
from click import ClickException
from pydantic import BaseModel, Field, root_validator
import tabulate
import yaml
from yaml.loader import SafeLoader

from anyscale.cli_logger import BlockLogger
from anyscale.cloud import get_cloud_id_and_name
from anyscale.cluster_compute import get_cluster_compute_from_name
from anyscale.conf import IDLE_TIMEOUT_DEFAULT_MINUTES
from anyscale.controllers.base_controller import BaseController
from anyscale.sdk.anyscale_client.models.cluster_computes_query import (
    ClusterComputesQuery,
)
from anyscale.sdk.anyscale_client.models.clustercompute_list_response import (
    ClustercomputeListResponse,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute import (
    CreateClusterCompute,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute_config import (
    CreateClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.text_query import TextQuery
from anyscale.util import get_endpoint
from anyscale.utils.entity_arg_utils import EntityType, IdBasedEntity, NameBasedEntity


log = BlockLogger()  # Anyscale CLI Logger


class CreateClusterComputeConfigModel(BaseModel):
    """
    Schema for ClusterClusterCompute that is accepted from the CLI. Supports specifying
    `cloud` name instead of `cloud_id`, and preprocesses so fields have the correct
    values to call POST `/ext/v0/cluster_computes`.
    """

    cloud_id: Optional[str] = Field(
        None, description="The ID of the Anyscale cloud to use for launching Clusters.",
    )
    cloud: Optional[str] = Field(
        None,
        description="The name of the Anyscale cloud to use for launching Clusters.",
    )
    max_workers: Optional[int] = Field(
        None, description="Desired limit on total running workers for this Cluster."
    )

    allowed_azs: Optional[List[str]] = Field(
        None,
        description='The availability zones that clusters are allowed to be launched in, e.g. "us-west-2a". If not specified, any AZ may be used.',
    )

    head_node_type: Any = Field(
        ..., description="Node configuration to use for the head node. ",
    )

    worker_node_types: List[Any] = Field(
        ..., description="A list of node types to use for worker nodes. "
    )

    aws: Optional[Any] = Field(
        None,
        description="Fields specific to AWS node types.",
        alias="aws_advanced_configurations_json",
    )

    gcp: Optional[Any] = Field(
        None,
        description="Fields specific to GCP node types.",
        alias="gcp_advanced_configurations_json",
    )

    azure: Optional[Any] = Field(
        None, description="Fields specific to Azure node types.",
    )

    maximum_uptime_minutes: Optional[int] = Field(
        None,
        description="If set to a positive number, Anyscale will terminate the cluster this many minutes after cluster start.",
    )

    idle_termination_minutes: int = Field(
        IDLE_TIMEOUT_DEFAULT_MINUTES,
        description=(
            "If set to a positive number, Anyscale will terminate the cluster this many minutes after the cluster is idle. "
            "Idle time is defined as the time during which a Cluster is not running a user command. "
            "Time spent running commands on Jupyter or ssh is still considered 'idle'. "
            "To disable, set this field to 0."
        ),
        ge=0,
    )

    @root_validator
    def fill_cloud_id(cls: Any, values: Any) -> Any:
        cloud_id, cloud = (
            values.get("cloud_id"),
            values.get("cloud"),
        )
        if cloud_id and cloud:
            raise click.ClickException(
                "Only one of `cloud_id` or `cloud` can be provided in the cluster compute config file. "
            )
        if cloud:
            cloud_id, _ = get_cloud_id_and_name(
                api_client=None, cloud_id=None, cloud_name=cloud
            )
            values["cloud_id"] = cloud_id
        elif not cloud_id:
            raise click.ClickException(
                "Please provide either `cloud_id` or `cloud` in the cluster compute config file. "
            )
        return values

    class Config:
        # This setting will cause pydantic (and fastAPI) to accept the field name as well as the alias
        # (eg. "cpu" or "CPU"). But note that openapi does not respect this setting.
        # Generated models will require the alias.
        allow_population_by_field_name = True


class ClusterComputeController(BaseController):
    """
    This controller powers functionalities related to Anyscale
    cluster compute configuration.
    """

    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.log.open_block("Output")

    def create(self, cluster_compute_file: IO[bytes], name: Optional[str]) -> None:
        """Builds a new cluster compute template
        If name is not provided, a default cluster-compute-name will be used and returned in the command output

        Information in output: Link to cluster compute in UI, cluster compute id
        """

        try:
            cluster_compute: Dict[str, Any] = yaml.load(
                cluster_compute_file, Loader=SafeLoader
            )
        except Exception as e:  # noqa: BLE001
            raise ClickException(f"Could not load cluster compute file: {e}")

        cluster_compute_config_model = CreateClusterComputeConfigModel(
            **cluster_compute
        )
        cluster_compute_config = CreateClusterComputeConfig(
            cloud_id=cluster_compute_config_model.cloud_id,
            max_workers=cluster_compute_config_model.max_workers,
            allowed_azs=cluster_compute_config_model.allowed_azs,
            region=None,
            head_node_type=cluster_compute_config_model.head_node_type,
            worker_node_types=cluster_compute_config_model.worker_node_types,
            aws_advanced_configurations_json=cluster_compute_config_model.aws,
            gcp_advanced_configurations_json=cluster_compute_config_model.gcp,
            maximum_uptime_minutes=cluster_compute_config_model.maximum_uptime_minutes,
            idle_termination_minutes=cluster_compute_config_model.idle_termination_minutes,
        )
        if name is None:
            name = f"cli-config-{datetime.now().isoformat()}"
        cluster_compute_response = self.anyscale_api_client.create_cluster_compute(
            CreateClusterCompute(name=name, config=cluster_compute_config)
        )
        created_cluster_compute = cluster_compute_response.result
        cluster_compute_id = created_cluster_compute.id
        cluster_compute_name = created_cluster_compute.name
        url = get_endpoint(f"/configurations/cluster-computes/{cluster_compute_id}")
        log.info(f"View this cluster compute at: {url}")
        log.info(f"Cluster compute id: {cluster_compute_id}")
        log.info(f"Cluster compute name: {cluster_compute_name}")

    def delete(
        self, cluster_compute_name: Optional[str], id: Optional[str]  # noqa: A002
    ) -> None:
        """
        TODO: shawnp. Remove once archive functionality is enabled on prod.
        Deletes the cluster compute with the given name or id.
        Exactly one of cluster_compute_name or id must be provided.
        """

        if int(bool(cluster_compute_name)) + int(bool(id)) != 1:
            raise ClickException(
                "Not deleted. Please provide exactly one of: cluster compute name, id."
            )

        if id:
            self.anyscale_api_client.get_cluster_compute(id)
            cluster_compute_id = id
        else:
            # find the cluster compute id from the name
            query = ClusterComputesQuery(name=TextQuery(equals=cluster_compute_name))
            query_response: ClustercomputeListResponse = self.anyscale_api_client.search_cluster_computes(
                query
            )
            cluster_results = query_response.results
            if len(cluster_results) != 1:
                raise ClickException(
                    f"Not deleted. No cluster compute template exists with the name {cluster_compute_name}."
                )
            cluster_compute_id = cluster_results[0].id
        self.anyscale_api_client.delete_cluster_compute(cluster_compute_id)
        log.info("Cluster compute deleted.")

    def archive(
        self, compute_config_entity: Union[IdBasedEntity, NameBasedEntity]
    ) -> None:
        """
        Archives the cluster compute with the given name or id.
        Exactly one of cluster_compute_name or id must be provided.
        """
        if compute_config_entity.type is EntityType.ID:
            compute_config = self.anyscale_api_client.get_compute_template(
                compute_config_entity.id
            ).result
            compute_config_id = compute_config_entity.id
            compute_config_name = compute_config.name
        else:
            compute_config = get_cluster_compute_from_name(
                compute_config_entity.name, self.api_client
            )
            compute_config_id = compute_config.id
            compute_config_name = compute_config.name
        self.api_client.archive_compute_template_api_v2_compute_templates_compute_template_id_archive_post(
            compute_config_id
        )
        log.info(f"Successfully archived compute config: {compute_config_name}.")

    def list(
        self,
        cluster_compute_name: Optional[str],
        cluster_compute_id: Optional[str],
        include_shared: bool,
        max_items: int,
    ) -> None:
        cluster_compute_list = []
        if cluster_compute_id:
            cluster_compute_list = [
                self.anyscale_api_client.get_cluster_compute(cluster_compute_id).result
            ]
        elif cluster_compute_name:
            cluster_compute_list = self.anyscale_api_client.search_cluster_computes(
                {"name": {"equals": cluster_compute_name}, "paging": {"count": 1}}
            ).results
        else:
            creator_id = (
                self.api_client.get_user_info_api_v2_userinfo_get().result.id
                if not include_shared
                else None
            )
            resp = self.anyscale_api_client.search_cluster_computes(
                {"creator_id": creator_id, "paging": {"count": 20}}
            )
            cluster_compute_list.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = (paging_token is not None) and (
                len(cluster_compute_list) < max_items
            )
            while has_more:
                resp = self.anyscale_api_client.search_cluster_computes(
                    {
                        "creator_id": creator_id,
                        "paging": {"count": 20, "paging_token": paging_token},
                    }
                )
                cluster_compute_list.extend(resp.results)
                paging_token = resp.metadata.next_paging_token
                has_more = (paging_token is not None) and (
                    len(cluster_compute_list) < max_items
                )
            cluster_compute_list = cluster_compute_list[:max_items]

        cluster_compute_table = [
            [
                cluster_compute.id,
                cluster_compute.name,
                self.anyscale_api_client.get_cloud(
                    cluster_compute.config.cloud_id
                ).result.name
                if cluster_compute.config.cloud_id
                else None,
                cluster_compute.last_modified_at.strftime("%m/%d/%Y, %H:%M:%S"),
                get_endpoint(f"configurations/cluster-computes/{cluster_compute.id}"),
            ]
            for cluster_compute in cluster_compute_list
        ]

        table = tabulate.tabulate(
            cluster_compute_table,
            headers=["ID", "NAME", "CLOUD", "LAST MODIFIED AT", "URL"],
            tablefmt="plain",
        )
        print(f"Cluster computes:\n{table}")

    def get(
        self, cluster_compute_name: Optional[str], cluster_compute_id: Optional[str],
    ) -> None:
        if (
            int(cluster_compute_name is not None) + int(cluster_compute_id is not None)
            != 1
        ):
            raise click.ClickException(
                "Please only provide one of `cluster-compute-name` or `--id`."
            )
        if cluster_compute_name:
            cluster_compute_id = get_cluster_compute_from_name(
                cluster_compute_name, self.api_client
            ).id
        compute_config = self.api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            cluster_compute_id
        ).result
        model_config = compute_config.config

        formatted_model_config = model_config.to_dict()
        formatted_model_config["archived_at"] = compute_config.archived_at
        # TODO(nikita): Improve formatting here
        pprint.pprint(formatted_model_config)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class AutoscalerReport(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'active_nodes': 'dict(str, int)',
        'pending_nodes': 'list[list[str]]',
        'pending_launches': 'dict(str, int)',
        'failed_nodes': 'list[list[str]]',
        'node_availability_summary': 'dict(str, object)'
    }

    attribute_map = {
        'active_nodes': 'activeNodes',
        'pending_nodes': 'pendingNodes',
        'pending_launches': 'pendingLaunches',
        'failed_nodes': 'failedNodes',
        'node_availability_summary': 'nodeAvailabilitySummary'
    }

    def __init__(self, active_nodes=None, pending_nodes=[], pending_launches=None, failed_nodes=[], node_availability_summary=None, local_vars_configuration=None):  # noqa: E501
        """AutoscalerReport - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._active_nodes = None
        self._pending_nodes = None
        self._pending_launches = None
        self._failed_nodes = None
        self._node_availability_summary = None
        self.discriminator = None

        if active_nodes is not None:
            self.active_nodes = active_nodes
        if pending_nodes is not None:
            self.pending_nodes = pending_nodes
        if pending_launches is not None:
            self.pending_launches = pending_launches
        if failed_nodes is not None:
            self.failed_nodes = failed_nodes
        if node_availability_summary is not None:
            self.node_availability_summary = node_availability_summary

    @property
    def active_nodes(self):
        """Gets the active_nodes of this AutoscalerReport.  # noqa: E501

        A dictionary of active nodes. The key is the node type and the value is the number of nodes of that type that are active.  # noqa: E501

        :return: The active_nodes of this AutoscalerReport.  # noqa: E501
        :rtype: dict(str, int)
        """
        return self._active_nodes

    @active_nodes.setter
    def active_nodes(self, active_nodes):
        """Sets the active_nodes of this AutoscalerReport.

        A dictionary of active nodes. The key is the node type and the value is the number of nodes of that type that are active.  # noqa: E501

        :param active_nodes: The active_nodes of this AutoscalerReport.  # noqa: E501
        :type: dict(str, int)
        """

        self._active_nodes = active_nodes

    @property
    def pending_nodes(self):
        """Gets the pending_nodes of this AutoscalerReport.  # noqa: E501

        A list of tuples where the first item is the ip address, the second item is the node type, and the third item is the status.  # noqa: E501

        :return: The pending_nodes of this AutoscalerReport.  # noqa: E501
        :rtype: list[list[str]]
        """
        return self._pending_nodes

    @pending_nodes.setter
    def pending_nodes(self, pending_nodes):
        """Sets the pending_nodes of this AutoscalerReport.

        A list of tuples where the first item is the ip address, the second item is the node type, and the third item is the status.  # noqa: E501

        :param pending_nodes: The pending_nodes of this AutoscalerReport.  # noqa: E501
        :type: list[list[str]]
        """

        self._pending_nodes = pending_nodes

    @property
    def pending_launches(self):
        """Gets the pending_launches of this AutoscalerReport.  # noqa: E501

        A dictionary of nodes pending to launch. The key is the node type and the value is the number of nodes of that type pending launch.  # noqa: E501

        :return: The pending_launches of this AutoscalerReport.  # noqa: E501
        :rtype: dict(str, int)
        """
        return self._pending_launches

    @pending_launches.setter
    def pending_launches(self, pending_launches):
        """Sets the pending_launches of this AutoscalerReport.

        A dictionary of nodes pending to launch. The key is the node type and the value is the number of nodes of that type pending launch.  # noqa: E501

        :param pending_launches: The pending_launches of this AutoscalerReport.  # noqa: E501
        :type: dict(str, int)
        """

        self._pending_launches = pending_launches

    @property
    def failed_nodes(self):
        """Gets the failed_nodes of this AutoscalerReport.  # noqa: E501

        A list of tuples where the first item is the ip address and the second item is the node type.  # noqa: E501

        :return: The failed_nodes of this AutoscalerReport.  # noqa: E501
        :rtype: list[list[str]]
        """
        return self._failed_nodes

    @failed_nodes.setter
    def failed_nodes(self, failed_nodes):
        """Sets the failed_nodes of this AutoscalerReport.

        A list of tuples where the first item is the ip address and the second item is the node type.  # noqa: E501

        :param failed_nodes: The failed_nodes of this AutoscalerReport.  # noqa: E501
        :type: list[list[str]]
        """

        self._failed_nodes = failed_nodes

    @property
    def node_availability_summary(self):
        """Gets the node_availability_summary of this AutoscalerReport.  # noqa: E501

        A mapping from node type to their availability. Introduced in Ray 2.1  # noqa: E501

        :return: The node_availability_summary of this AutoscalerReport.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._node_availability_summary

    @node_availability_summary.setter
    def node_availability_summary(self, node_availability_summary):
        """Sets the node_availability_summary of this AutoscalerReport.

        A mapping from node type to their availability. Introduced in Ray 2.1  # noqa: E501

        :param node_availability_summary: The node_availability_summary of this AutoscalerReport.  # noqa: E501
        :type: dict(str, object)
        """

        self._node_availability_summary = node_availability_summary

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AutoscalerReport):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, AutoscalerReport):
            return True

        return self.to_dict() != other.to_dict()



# NOTE: The logic in this file/module is only run from within an Anyscale workspace
import datetime
import json
import logging
import os
import re
import shutil
import subprocess
import tarfile
import tempfile
import time
from typing import Any, Dict, Optional, Tuple

from anyscale.utils.imports.all import try_import_ray


logger = logging.getLogger(__name__)

# Built-in configs.
WORKING_DIR = os.environ.get("ANYSCALE_WORKING_DIR", "/home/ray")
# These files are saved per workspace.
CONFIGS_TO_SAVE = [
    ".bash_history",
    ".zsh_history",
]
# These files are saved per user.
CONFIGS_TO_REPLICATE = [
    ".workspacerc",
    ".gitconfig",
    ".vimrc",
    ".tmux.conf",
]

# Default cluster storage dir.
CLUSTER_STORAGE_DIR = "/mnt/cluster_storage"

# Default VSCode server dir.
VSCODE_DESKTOP_SERVER_DIR = "/home/ray/.vscode-server"

# Experimental configs propagated from the cluster config service.
EFS_IP = os.environ.get("ANYSCALE_EXPERIMENTAL_EFS_IP", "")
EFS_ROOT = os.environ.get("ANYSCALE_EXPERIMENTAL_EFS_ROOT", "")
EFS_TYPE = os.environ.get("ANYSCALE_EXPERIMENTAL_EFS_TYPE", "")
# Override with container-defined IP.
CUSTOM_EFS_IP = os.environ.get("CUSTOM_EFS_IP", "")
CUSTOM_EFS_ROOT = os.environ.get("CUSTOM_EFS_ROOT", "")
IDLE_TERMINATION_DIR = os.environ.get(
    "IDLE_TERMINATION_DIR", "/tmp/anyscale/idle_termination_reports"
)
if CUSTOM_EFS_IP:
    EFS_IP = CUSTOM_EFS_IP

if CUSTOM_EFS_ROOT:
    EFS_ROOT = CUSTOM_EFS_ROOT

WORKSPACE_ID = os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID", "")
USERNAME = os.environ.get("ANYSCALE_EXPERIMENTAL_USERNAME", "unknown_user")
BASE_SNAPSHOT = os.environ.get("ANYSCALE_EXPERIMENTAL_BASE_SNAPSHOT") or None

# GitHub URI to use as the template for the new workspace. This is the URI you see
# when browsing the desired subdir in the web UI, e.g.:
#
# https://github.com/ray-project/ray/tree/master/release/ml_user_tests
#
WORKSPACE_TEMPLATE = os.environ.get("ANYSCALE_WORKSPACE_TEMPLATE") or None

# Other debug configs.
SNAPSHOT_INTERVAL = int(os.environ.get("WORKSPACE_SNAPSHOT_INTERVAL", 300))
EFS_WORKSPACE_DIR = os.environ.get("EFS_WORKSPACE_DIR", "/efs/workspaces")
EFS_JOB_DIR = os.environ.get("EFS_WORKSPACE_DIR", "/efs/jobs")
EFS_OBJECTS_DIR = os.environ.get("EFS_OBJECTS_DIR", "/efs/workspaces/shared_objects")
EFS_CREDS_DIR = os.environ.get("EFS_CREDS_DIR", "/efs/generated_credentials")
RAY_ML_DEV = bool(os.environ.get("RAY_ML_DEV"))
AUTOGC = bool(os.environ.get("AUTOGC_SNAPSHOTS", True))
SNAPSHOT_RETENTION_COUNT = max(int(os.environ.get("SNAPSHOT_RETENTION_COUNT", 5)), 1)
SNAPSHOT_RETENTION_HOURS = max(int(os.environ.get("SNAPSHOT_RETENTION_HOURS", 1)), 1)
SKIP_VSCODE_DESKTOP_SETUP = bool(os.environ.get("SKIP_VSCODE_DESKTOP_SETUP", False))

# This causes workspaces to use the $CWD as the runtime env working dir instead of
# always the workspace root.
RELATIVE_WORKING_DIR = bool(os.environ.get("RELATIVE_WORKING_DIR", True))


def optimize_git_repo(directory: str, shared_repo: str) -> None:
    """Optimize the space usage of a git repo by syncing objects to a shared repo.

    Any objects in the source repo will be replicated to the shared repo, and then
    deleted from the source repo. The source repo is setup to reference objects in
    the shared repo via the `.git/objects/info/alternates` mechanism.

    Args:
        directory: The directory to optimize.
        shared_repo: The path that should be used to hold shared objects. This path
            will be created if it doesn't already exist. Multiple checkouts of the
            same repo can share the objects stored in the shared repo.
    """
    start = time.time()
    objects_path = f"{directory}/.git/objects"
    if os.path.exists(objects_path):
        if not os.path.exists(shared_repo):
            os.makedirs(os.path.dirname(shared_repo), exist_ok=True)
            # TODO(ekl) it's faster to do a copy of just the objects dir, but it seems
            # we need to git clone in order for alternates to be recognized as valid.
            subprocess.check_call(
                f"git clone --bare {directory}/ {shared_repo}/", shell=True,
            )
        shared_objects_dir = os.path.join(shared_repo, "objects")
        subprocess.check_call(
            f"rsync -a {objects_path}/ {shared_objects_dir}/", shell=True
        )
        subprocess.check_call("rm -rf {}".format(objects_path), shell=True)  # noqa
        os.makedirs(os.path.join(objects_path, "info"), exist_ok=True)
        with open(os.path.join(objects_path, "info/alternates"), "w") as f:
            f.write(f"{shared_objects_dir}\n")
    logger.info(
        "Synced git objects for {} to {} in {}s.".format(
            directory, shared_repo, time.time() - start
        )
    )


def create_snapshot_zip(directory: str, auto: bool) -> str:
    """Create a snapshot of the given directory.

    The snapshot will include all git tracked files as well as unstaged
    (but otherwise trackable) files. It will also include the full
    contents of the `.git` folder. To optimize the disk space usage of
    snapshots, call `optimize_git_repo` on the repo directory prior to
    calling `create_snapshot_zip`.

    Args:
        directory: Path of the directory to snapshot.

    Returns:
        Path of a .zip file that contains the snapshot files.
    """

    start = time.time()
    orig = os.path.abspath(os.curdir)
    prefix = "snapshot_{}_".format(
        datetime.datetime.now(datetime.timezone.utc).isoformat()
    )
    if auto:
        prefix += "auto_"
    target = tempfile.mktemp(suffix=".zip", prefix=prefix)
    is_git_workspace = os.path.exists(os.path.join(WORKING_DIR, ".git"))
    is_within_workspace = os.path.abspath(directory).startswith(
        os.path.abspath(WORKING_DIR)
    )
    try:
        os.chdir(directory)
        if is_git_workspace and is_within_workspace:
            subprocess.check_call(
                "(git ls-files -co --exclude-standard || true; find .git || true) | "
                f"zip --symlinks -@ -0 -q {target}",
                shell=True,
            )
        else:
            for child in os.listdir("."):
                if os.path.exists(os.path.join(child, ".git")):
                    raise ValueError(
                        f"Git repo detected in sub-directory {child}. Please ensure "
                        "that your git repo is cloned in the top-level workspace "
                        f"directory with 'git clone <repo> .' at {directory}."
                    )
            subprocess.check_call(
                f"find . | zip --symlinks -@ -0 -q {target}", shell=True,
            )
    finally:
        os.chdir(orig)

    assert os.path.exists(target), target
    logger.info(
        "Created snapshot for {} at {} of size {} in {}s.".format(
            directory, target, os.path.getsize(target), time.time() - start
        )
    )
    return target


def unpack_snapshot_zip(zip_path: str, directory: str) -> None:
    """Unpack a snapshot to the given directory.

    Args:
        zip_path: Path of the zip returned by create_snapshot_zip.
        directory: Output directory to unpack the zip into.
    """

    start = time.time()
    os.makedirs(directory, exist_ok=True)
    subprocess.check_call(f"unzip -X -o -q {zip_path} -d {directory}", shell=True)
    logger.info(
        "Unpacked snapshot {} to {} in {}s.".format(
            zip_path, directory, time.time() - start
        )
    )


def unpack_github_dir(github_tree_uri: str, out_dir: str) -> None:
    """Download a GitHub tree URI for a directory.

    Args:
        github_tree_uri: URL for a directory, for example:
            "https://github.com/ray-project/ray/tree/master/release/ml_user_tests"
        out_dir: The output dir to put files, which must be empty.
    """

    if "github.com" not in github_tree_uri:
        raise ValueError("Only GitHub URLs are supported by load_package().")

    # Normalize dir URIs.
    if github_tree_uri.endswith("/"):
        github_tree_uri = github_tree_uri[:-1]

    URL_FORMAT = ".*github.com/([^/]*)/([^/]*)/tree/([^/]*)/(.*)"
    match = re.match(URL_FORMAT, github_tree_uri)
    if not match:
        raise ValueError(f"GitHub URL must be of format {URL_FORMAT}")
    gh_user = match.group(1)
    gh_repo = match.group(2)
    gh_branch = match.group(3)
    gh_subdir = match.group(4)

    logger.info(
        f"Attempting download, gh_user={gh_user}, gh_repo={gh_repo}, "
        f"gh_branch={gh_branch}, gh_subdir={gh_subdir}"
    )

    tmp = tempfile.mktemp(prefix=f"github_{gh_repo}", suffix=".tar.gz")
    subprocess.check_call(
        [
            "curl",
            "--fail",
            "-L",
            f"https://github.com/{gh_user}/{gh_repo}/tarball/{gh_branch}",
            "--output",
            tmp,
        ]
    )
    logger.info(f"Downloaded tarball to {tmp}")

    # The directory we want to extract is a subdir within the zip, e.g.,
    # ray-project-fc17342/release/ml_user_tests/ray-lightning
    top_dir = tarfile.open(tmp).getnames()[0]
    tar_path = top_dir + "/" + gh_subdir
    # Calculate the depth to the subdir and strip those leading components.
    strip = tar_path.count("/") + 1
    tar_command = [
        "tar",
        "xzf",
        tmp,
        "-C",
        out_dir,
        tar_path,
        f"--strip-components={strip}",
    ]
    logger.info(f"Extract command is {tar_command}")
    subprocess.check_call(tar_command)


def compute_content_hash(zip_path: str) -> bytes:
    """Return the md5 hash of a given zipfile on disk."""
    md5 = subprocess.check_output(
        f"unzip -p {zip_path} | md5sum -b | cut -f1 -d ' '", shell=True
    )
    md5 = md5.strip()
    return md5


def get_or_create_snapshot_zip(directory: str, auto: bool) -> str:
    """Create a snapshot zip, or return the last snapshot if unchanged.

    A corresponding .md5 file is created alongside the snapshot zip.
    """
    new_zip = create_snapshot_zip(directory, auto)
    new_hash = compute_content_hash(new_zip)
    # Ignore the base snapshot in auto save mode. This means we will always generate
    # a new snapshot within the autosave interval, allowing the base snapshot to be
    # safely garbage collected.
    old_zip = find_latest(ignore_base_snapshot=auto)
    if old_zip:
        try:
            with open(old_zip + ".md5", "rb") as f:
                old_hash: Optional[bytes] = f.read().strip()
        except Exception:  # noqa: BLE001
            logger.warning("Failed to read md5 file")
            old_hash = None
    else:
        old_hash = None
    logger.info(f"Content hashes {old_hash!r} vs {new_hash!r}")
    if old_hash == new_hash:
        logger.info("Content hash unchanged, not saving new snapshot.")
        os.unlink(new_zip)
        assert old_zip is not None
        return old_zip
    else:
        with open(new_zip + ".md5", "wb") as f:
            f.write(new_hash)
        return new_zip


def do_snapshot(auto: bool = False):
    """Command to create a snapshot within an Anyscale workspace.

    Can be run via `python -m anyscale.snapshot_util snapshot`.
    """

    try:
        workspace_dir = os.path.join(EFS_WORKSPACE_DIR, WORKSPACE_ID)
        snapshot_dir = os.path.join(workspace_dir, "snapshots")

        is_empty = len(os.listdir(WORKING_DIR)) == 0

        # directory is empty, no need to snapshot
        if is_empty:
            return

        # TODO(ekl) should we isolate the objects by workspace or repo?
        optimize_git_repo(WORKING_DIR, EFS_OBJECTS_DIR)
        zip = get_or_create_snapshot_zip(WORKING_DIR, auto)  # noqa: A001

        # If the zip was already on EFS, we're done.
        if zip.startswith(snapshot_dir):
            return

        # Otherwise, move the zip into EFS along with its md5 file.
        os.makedirs(snapshot_dir, exist_ok=True)
        shutil.move(zip, os.path.join(snapshot_dir, os.path.basename(zip)))
        shutil.move(
            zip + ".md5", os.path.join(snapshot_dir, os.path.basename(zip) + ".md5")
        )
        for config in CONFIGS_TO_SAVE:
            source = os.path.join("/home/ray", config)
            if os.path.exists(source):
                shutil.copy(source, os.path.join(workspace_dir, config))

        # report successful idle termination status
        report_idle_termination_activity()

    except Exception:
        logger.exception("Failed to create a snapshot")
        # report error idle termination status
        report_idle_termination_error("Failed to create a snapshot")

    if AUTOGC:
        gc_snapshots()


def report_idle_termination_activity():
    if os.path.exists(IDLE_TERMINATION_DIR):
        with open(f"{IDLE_TERMINATION_DIR}/workspace.json", "w") as f:
            f.write(json.dumps({"last_activity_timestamp": time.time()}))
    else:
        logger.warning(
            f"Did not report workspace idle termination status because {IDLE_TERMINATION_DIR} is not mounted."
        )


def report_idle_termination_error(err: str):
    content = {"error": err}
    if os.path.exists(IDLE_TERMINATION_DIR):
        with open(f"{IDLE_TERMINATION_DIR}/workspace.json", "w") as f:
            f.write(json.dumps(content))
    else:
        logger.warning(
            f"Did not report workspace idle termination status because {IDLE_TERMINATION_DIR} is not mounted."
        )


def gc_snapshots() -> None:
    """Garbage collect snapshots older than X hours.

    This is safe since when we clone a workspace, the autosave loop will generate a
    new snapshot, so the clone will be decoupled from the original.
    """
    workspace_dir = os.path.join(EFS_WORKSPACE_DIR, WORKSPACE_ID)
    snapshot_dir = os.path.join(workspace_dir, "snapshots")
    if not os.path.exists(snapshot_dir):
        return
    snapshots = sorted([x for x in os.listdir(snapshot_dir) if x.endswith(".zip")])
    # Never GC the latest SNAPSHOT_RETENTION_COUNT snapshot.
    snapshots = snapshots[:-SNAPSHOT_RETENTION_COUNT]
    horizon = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
        hours=SNAPSHOT_RETENTION_HOURS
    )

    logger.debug(
        f"Snapshot retention count: {SNAPSHOT_RETENTION_COUNT}, snapshot retention hours: {SNAPSHOT_RETENTION_HOURS}"
    )

    prefix = f"snapshot_{horizon.isoformat()}_"
    # GC snapshots older than SNAPSHOT_RETENTION_HOURS hour.
    for snapshot in snapshots:
        if snapshot < prefix:
            full_path = os.path.join(snapshot_dir, snapshot)
            md5 = full_path + ".md5"
            logger.info(f"Deleting old snapshot {full_path}")
            try:
                os.unlink(full_path)
                os.unlink(md5)
            except Exception:
                logger.exception("Failed to delete snapshot")


def find_latest(ignore_base_snapshot: bool) -> Optional[str]:
    """Return path to latest .zip snapshot, if it exists."""
    workspace_dir = os.path.join(EFS_WORKSPACE_DIR, WORKSPACE_ID)
    snapshot_dir = os.path.join(workspace_dir, "snapshots")
    if not os.path.exists(snapshot_dir):
        if ignore_base_snapshot:
            return None
        else:
            return find_base_snapshot()
    snapshots = sorted([x for x in os.listdir(snapshot_dir) if x.endswith(".zip")])
    if not snapshots:
        return find_base_snapshot()
    return os.path.join(snapshot_dir, snapshots[-1])


def find_base_snapshot() -> Optional[str]:  # noqa: PLR0911
    if not BASE_SNAPSHOT:
        return None
    try:
        base_data = json.loads(BASE_SNAPSHOT)
    except Exception:
        logger.exception("Failed to parse base snapshot info")
        return None

    # Jobs snapshot
    if "from_job" in base_data:
        job_id = base_data["from_job"]["job_id"]
        if not job_id:
            logger.info("Invalid base snapshot, no job id")
            return None
        logger.info(f"Base snapshot from job {job_id}")
        return os.path.join(EFS_JOB_DIR, job_id, "working_dir.zip")

    # Workspace snapshot
    if "from_workspace" in base_data:
        workspace_id = base_data["from_workspace"]["workspace_id"]
        iso_time = base_data["from_workspace"]["iso_time"]
        if not workspace_id or not iso_time:
            logger.info("Invalid base snapshot, no workspace id or time")
            return None
        logger.info(f"Base snapshot from workspace {workspace_id} and {iso_time}")
        snapshot_dir = os.path.join(EFS_WORKSPACE_DIR, workspace_id, "snapshots")
        snapshot_time = f"snapshot_{iso_time}_"
        if not os.path.exists(snapshot_dir):
            return None
        # Find the latest matching snapshot before the time stamp
        snapshots = sorted(
            [
                x
                for x in os.listdir(snapshot_dir)
                if x.endswith(".zip") and x < snapshot_time
            ]
        )
        if not snapshots:
            return None
        return os.path.join(snapshot_dir, snapshots[-1])

    return None


def restore_latest():
    """Command to restore the latest snapshot within an Anyscale workspace.

    Can be run via `python -m anyscale.snapshot_util restore`.
    """
    latest = find_latest(ignore_base_snapshot=False)
    logger.info(f"Latest snapshot found was {latest}")
    if not latest:
        # No snapshot found: fallback to template or blank workspace.
        if WORKSPACE_TEMPLATE:
            unpack_github_dir(WORKSPACE_TEMPLATE, WORKING_DIR)
        return
    workspace_dir = os.path.join(EFS_WORKSPACE_DIR, WORKSPACE_ID)
    unpack_snapshot_zip(latest, WORKING_DIR)
    for config in CONFIGS_TO_SAVE:
        saved = os.path.join(workspace_dir, config)
        if os.path.exists(saved):
            shutil.copy(saved, os.path.join("/home/ray", config))


# Copied from https://stackoverflow.com/a/50864227
def _split_s3_path(s3_path: str) -> Tuple[str, str]:
    path_parts = s3_path.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)
    return bucket, key


def checkpoint_job(job_id, runtime_env: Dict[str, Any]) -> Dict[str, Any]:
    """Checkpoint the runtime environment and working directory of a job.

    This function will modify runtime_env to point to the new working
    directory on EFS and return the modified runtime_env.
    """
    # For some reason, in a job Ray first calls the env_hook with runtime_env = None
    # and then a second time with the proper runtime_env -- do nothing in the former case.
    if not runtime_env:
        return runtime_env
    dest_dir = os.path.join(EFS_JOB_DIR, job_id)
    os.makedirs(dest_dir, exist_ok=True)
    if "working_dir" in runtime_env and runtime_env["working_dir"].endswith(".zip"):
        # We're a job, also save a replica of the zip in EFS so the job can be cloned
        # as a workspace at a later time.
        working_dir = os.path.join(dest_dir, "working_dir.zip")

        # We need a custom path for s3 urls because urllib.request.urlretrieve doesn't work for s3 urls
        if runtime_env["working_dir"].startswith("s3://"):
            import boto3

            bucket, key = _split_s3_path(runtime_env["working_dir"])
            # boto3 client initialized inside clusters already have AWS credentials set
            s3_client = boto3.client("s3")
            s3_client.download_file(bucket, key, working_dir)
        else:
            import urllib.request

            urllib.request.urlretrieve(runtime_env["working_dir"], working_dir)
        runtime_env["working_dir"] = working_dir
    # Save the runtime_env to be used later.
    with open(os.path.join(dest_dir, "runtime_env.json"), "w") as f:
        f.write(json.dumps(runtime_env))
    return runtime_env


def setup_ml_dev(runtime_env):
    """Env hook for Ray ML development.

    This enables development for Ray ML libraries, assuming the working dir is the
    entire Ray repo, by replicating library changes to all nodes in the cluster via
    runtime_env py_modules.

    To enable this hook, set RAY_ML_DEV=1.
    """
    if not runtime_env:
        runtime_env = {}
    ray = try_import_ray()

    sys_ray_module = os.path.dirname(ray.__file__)
    local_ray_module = os.path.join(WORKING_DIR, "python/ray")
    if not os.path.exists(local_ray_module):
        logger.info("RAY_ML_DEV was set, but could not find the local ray module.")
        return runtime_env
    tmp_module = "/tmp/ray_tmp_module/ray"
    shutil.rmtree(tmp_module, ignore_errors=True)
    shutil.copytree(sys_ray_module, tmp_module)
    # TODO(ekl) keep this in sync with setup-dev.py
    LIB_DIRS = [
        "rllib",
        "air",
        "tune",
        "serve",
        "train",
        "data",
        "experimental",
        "util",
        "workflow",
        "cloudpickle",
        "_private",
        "internal",
        "node.py",
        "cluster_utils.py",
        "ray_constants.py",
    ]
    for lib_dir in LIB_DIRS:
        src = os.path.join(local_ray_module, lib_dir)
        dst = os.path.join(tmp_module, lib_dir)
        logger.info(f"Copying files from {src} to {dst}.")
        if os.path.isdir(src):
            shutil.rmtree(dst)
            shutil.copytree(os.path.join(local_ray_module, lib_dir), dst)
        elif os.path.exists(src):
            shutil.copy(src, dst)
        else:
            logger.info(f"Did not find {src}")
    if "py_modules" not in runtime_env:
        runtime_env["py_modules"] = []
    runtime_env["py_modules"].append(tmp_module)
    return runtime_env


def env_hook(runtime_env) -> Dict[str, Any]:
    """Env hook for including the working dir in the runtime_env by default.

    This should be set as `RAY_RUNTIME_ENV_HOOK=anyscale.snapshot_util.env_hook`.
    """
    if not runtime_env:
        runtime_env = {}

    # If EFS fails to mount, we don't update the working_dir to EFS
    if not os.path.ismount("/efs"):
        return runtime_env

    if "ANYSCALE_EXPERIMENTAL_INITIAL_JOB_ID" in os.environ:
        return checkpoint_job(
            os.environ["ANYSCALE_EXPERIMENTAL_INITIAL_JOB_ID"], runtime_env
        )

    # Only fill in the working_dir if it's non-empty (to avoid empty zip errors).
    if not runtime_env.get("working_dir") and os.listdir(WORKING_DIR):
        optimize_git_repo(WORKING_DIR, EFS_OBJECTS_DIR)
        if RELATIVE_WORKING_DIR:
            zipfile = create_snapshot_zip(".", auto=False)
        else:
            zipfile = create_snapshot_zip(WORKING_DIR, auto=False)
        # Move the zip file to a consistent path so that we don't leak zip files as
        # jobs are run over time. This isn't thread safe: we assume one job at a time.
        # Note that the zip file isn't needed after the job starts successfully.
        final_path = "/tmp/ray_latest_runtime_env.zip"
        shutil.move(zipfile, final_path)
        runtime_env["working_dir"] = final_path
    if RAY_ML_DEV:
        runtime_env = setup_ml_dev(runtime_env)
    env_vars = runtime_env.get("env_vars", {})
    env_vars.update(_load_env_vars())
    # This if condition is a workaround for
    # https://github.com/anyscale/product/issues/11679
    if env_vars:
        runtime_env["env_vars"] = env_vars
    logger.info(f"Updated runtime env to {runtime_env}")
    return runtime_env


def job_hook(**kwargs):
    """Job hook to tell users to use normal commands or anyscale job submit.

    This should be set as `RAY_JOB_SUBMIT_HOOK=anyscale.snapshot_util.job_hook`.
    """
    cmd = " ".join(kwargs["entrypoint"])
    raise ValueError(
        f"`ray job submit` is not needed within a workspace. Run `{cmd}` "
        f"directly to execute the job in the current cluster. To run on a new "
        f"cluster, use `anyscale job submit -- {cmd}`."
    )


def setup_credentials():
    """Command to create SSH credentials for the workspace.

    We generate unique Anyscale SSH keys for each username. This call will inject
    the key for the current user into the workspace.
    """
    private_key = os.path.join(EFS_CREDS_DIR, USERNAME, "id_rsa")
    public_key = os.path.join(EFS_CREDS_DIR, USERNAME, "id_rsa.pub")
    os.path.join(EFS_CREDS_DIR, USERNAME, "env_vars.json")

    # setup ssh directory commands
    setup_ssh_directory_cmds = [
        "mkdir -p /home/ray/.ssh",
        # incase the directory is already created by docker, make sure we can write to it.
        "sudo chown ray /home/ray/.ssh",
    ]

    for cmd in setup_ssh_directory_cmds:
        try:
            subprocess.check_call(cmd, shell=True)
        except Exception:
            logger.exception(f"Error running {cmd}")

    if os.path.exists(private_key):
        # Copy down from EFS.
        shutil.copy(private_key, "/home/ray/.ssh/id_rsa")
        shutil.copy(public_key, "/home/ray/.ssh/id_rsa.pub")
    else:
        # Copy up to EFS.
        subprocess.check_call(
            "echo y | ssh-keygen -t rsa -f /home/ray/.ssh/id_rsa -N ''", shell=True
        )
        os.makedirs(os.path.dirname(private_key), exist_ok=True)
        shutil.copy("/home/ray/.ssh/id_rsa", private_key)
        shutil.copy("/home/ray/.ssh/id_rsa.pub", public_key)

    # Also put stored user env vars in the bashrc.
    env_vars = _load_env_vars()
    with open("/home/ray/.bashrc", "a") as bashrc:
        bashrc.write("\n")
        for key, value in env_vars.items():
            bashrc.write(f"export {key}='{value}'\n")


def _load_env_vars():
    env_store = os.path.join(EFS_CREDS_DIR, USERNAME, "env_vars.json")
    env_vars = {}
    if os.path.exists(env_store):
        try:
            with open(env_store) as f:
                env_vars = json.loads(f.read())
        except Exception:
            logger.exception("Error parsing env vars")
    return env_vars


def _save_env_vars(env_vars: dict):
    env_store = os.path.join(EFS_CREDS_DIR, USERNAME, "env_vars.json")
    with open(env_store, "w") as f:
        f.write(json.dumps(env_vars))
    print(f"Stored keys: {list(env_vars)}")


def get_env_var(key: str):
    env_vars = _load_env_vars()
    if key not in env_vars:
        raise KeyError("No value was put for", key)
    print(f"{key}={env_vars[key]}")


def del_env_var(key: str):
    env_vars = _load_env_vars()
    if key in env_vars:
        del env_vars[key]
    _save_env_vars(env_vars)


def put_env_var(key: str, value: str) -> None:
    env_vars = _load_env_vars()
    env_vars[key] = value
    _save_env_vars(env_vars)


def setup_nfs():
    """Setup EFS or FileStore mounts in the container."""
    EFS_MOUNT = f"sudo mount -t nfs4 -o nfsvers=4.1,rsize=1048576,wsize=1048576,hard,timeo=600,retrans=2,noresvport {EFS_IP}:/ /efs"
    FILE_STORE_MOUNT = f"sudo mount -o rsize=1048576,wsize=1048576,hard,timeo=600,retrans=2,nolock {EFS_IP}:/{EFS_ROOT} /efs"
    COMMANDS = [
        "sudo mkdir -p /efs",
        EFS_MOUNT if EFS_TYPE == "EFS" else FILE_STORE_MOUNT,
        "sudo mkdir -p /efs",
        "sudo chown ray /efs",
        f"mkdir -p /efs/workspaces/{WORKSPACE_ID}/cluster_storage",
        f"sudo ln -sf /efs/workspaces/{WORKSPACE_ID}/cluster_storage {CLUSTER_STORAGE_DIR}",
        f"mkdir -p /efs/users/{USERNAME}",
        f"sudo ln -sf /efs/users/{USERNAME} /mnt/user_storage",
        "mkdir -p /efs/shared_storage",
        "sudo ln -sf /efs/shared_storage /mnt/shared_storage",
    ]
    for config in CONFIGS_TO_REPLICATE:
        COMMANDS.append(f"ln -sf /mnt/user_storage/{config} /home/ray/{config}")
    for cmd in COMMANDS:
        try:
            subprocess.check_call(cmd, shell=True)
        except Exception:
            logger.exception(f"Error running {cmd}")
            raise

    # on first load, report an activity to signal healthy start
    report_idle_termination_activity()


def setup_vscode_desktop():
    """Setup vscode desktop extension directory."""

    if SKIP_VSCODE_DESKTOP_SETUP:
        return

    local_folder = VSCODE_DESKTOP_SERVER_DIR
    remote_folder = os.path.join(CLUSTER_STORAGE_DIR, "vscode_desktop")

    # create the local folder
    os.makedirs(local_folder, exist_ok=True)

    # create the remote sub folders and link them
    sub_folders = ["data", "extensions"]
    for folder in sub_folders:
        directory = f"{remote_folder}/{folder}"
        os.makedirs(directory, exist_ok=True)
        os.symlink(directory, f"{local_folder}/{folder}")


def setup_container(ray_params: Any, is_head: bool):  # noqa: ARG001, PLR0912
    """Setup the container synchronously prior to Ray start.

    This handles (1) mounting network storage, (2) restoring workspace data, and
    (3) restoring credentials. This is intended to be triggered via the Ray start hook,
    i.e., ``RAY_START_HOOK=anyscale.snapshot_util.setup_container``.
    """
    if os.path.exists("/tmp/initialized"):
        logger.info("Init previously completed, skipping.")
        return
    if not EFS_IP:
        logger.info("No EFS IP configured, skipping workspace container setup.")
    else:
        try:
            setup_nfs()
        except Exception:
            logger.exception("Failed to setup NFS")
            raise
        if not WORKSPACE_ID:
            logger.info("No workspace id configured, skipping workspace restore.")
        elif not is_head:
            logger.info("Not head node, skipping workspace restore.")
        else:
            try:
                restore_latest()
            except Exception:
                logger.exception("Failed to restore workspace")
            try:
                setup_credentials()
            except Exception:
                logger.exception("Failed to setup SSH credentials")
            try:
                setup_vscode_desktop()
            except Exception:
                logger.exception("Failed to setup VSCode desktop")
    with open("/tmp/initialized", "w") as f:
        f.write("ok")


def autosnapshot_loop():
    if not WORKSPACE_ID:
        logger.info("Workspaces disabled.")
        return
    logger.info(f"Started autosnapshot loop with interval {SNAPSHOT_INTERVAL}")
    while True:
        time.sleep(SNAPSHOT_INTERVAL)
        do_snapshot(auto=True)


def post_build():
    CMDS = [
        "pip install -U https://s3-us-west-2.amazonaws.com/ray-wheels/latest/ray-3.0.0.dev0-cp38-cp38-manylinux2014_x86_64.whl",
        # TODO: remove this after next image release.
        "echo 'PROMPT_COMMAND=\"history -a\"' >> /home/ray/.bashrc",
        "echo '[ -e ~/.workspacerc ] && source ~/.workspacerc' >> /home/ray/.bashrc",
        f"echo 'export PATH={CLUSTER_STORAGE_DIR}/pypi/bin:$PATH' >> /home/ray/.bashrc",
    ]
    for cmd in CMDS:
        print("Executing", cmd)
        subprocess.check_call(cmd, shell=True)


if __name__ == "__main__":
    import sys

    # Since this is also a main script, we need to set the log level here
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(os.environ.get("ANYSCALE_LOGLEVEL", "WARN"))

    if sys.argv[1] == "snapshot":
        do_snapshot()
    elif sys.argv[1] == "post_build":
        post_build()
    elif sys.argv[1] == "autosnapshot":
        autosnapshot_loop()
    elif sys.argv[1] == "restore":
        restore_latest()
    elif sys.argv[1] == "gc":
        gc_snapshots()
    elif sys.argv[1] == "put_env":
        put_env_var(sys.argv[2], sys.argv[3])
    elif sys.argv[1] == "get_env":
        get_env_var(sys.argv[2])
    elif sys.argv[1] == "del_env":
        del_env_var(sys.argv[2])
    elif sys.argv[1] == "setup_credentials":
        setup_credentials()
    elif sys.argv[1] == "setup_container":
        setup_container(None, True)
    else:
        print("unknown arg")



import json
import os
from typing import Optional
from unittest.mock import call, Mock, mock_open, patch

import click
from pydantic import ValidationError
import pytest

from anyscale.client.openapi_client.models.apply_production_service_v2_api_model import (
    ApplyProductionServiceV2APIModel,
)
from anyscale.client.openapi_client.models.cloud_providers import CloudProviders
from anyscale.client.openapi_client.models.create_production_service import (
    CreateProductionService,
)
from anyscale.client.openapi_client.models.decoratedlistserviceapimodel_list_response import (
    DecoratedlistserviceapimodelListResponse,
)
from anyscale.client.openapi_client.models.decoratedsession_response import (
    DecoratedsessionResponse,
)
from anyscale.client.openapi_client.models.list_response_metadata import (
    ListResponseMetadata,
)
from anyscale.client.openapi_client.models.production_job_config import (
    ProductionJobConfig,
)
from anyscale.controllers.service_controller import ServiceConfig, ServiceController


TEST_SERVICE_ID = "test-service-id"
TEST_SERVICE_NAME = "test-service-name"


def test_service_config_model_with_entrypoint():
    config_dict = {
        "entrypoint": "python test.py",
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
        "healthcheck_url": "test.url",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.ray_serve_config is None


def test_service_config_model_with_ray_serve_config():
    runtime_env = {
        "pip": ["requests"],
        "working_dir": "https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip",
    }
    config_dict = {
        "ray_serve_config": {
            "import_path": "serve_hello:entrypoint",
            "runtime_env": runtime_env,
        },
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.entrypoint is None
        assert service_config.runtime_env is None
        assert service_config.ray_serve_config["runtime_env"] == runtime_env


def test_service_config_model_with_ray_serve_config_working_dir():
    runtime_env = {"pip": ["requests"], "working_dir": "."}
    config_dict = {
        "ray_serve_config": {
            "import_path": "serve_hello:entrypoint",
            "runtime_env": runtime_env,
        },
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert service_config.entrypoint is None
        assert service_config.ray_serve_config["runtime_env"] == runtime_env
        assert service_config.ray_gcs_external_storage_config is None


def test_service_config_model_with_ray_gcs_external_storage_config():
    config_dict = {
        "ray_serve_config": {"runtime_env": {"pip": ["requests"], "working_dir": "."}},
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
        "ray_gcs_external_storage_config": {"address": "addr:port",},
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)
        assert (
            service_config.ray_gcs_external_storage_config
            == config_dict["ray_gcs_external_storage_config"]
        )
        assert service_config.runtime_env is None


@pytest.mark.parametrize("use_default_project", [True, False])
@pytest.mark.parametrize("access", ["public", "private"])
@pytest.mark.parametrize("maximum_uptime_minutes", [60, None])
def test_update_service(
    mock_auth_api_client,
    use_default_project: bool,
    access: str,
    maximum_uptime_minutes: Optional[int],
) -> None:
    config_dict = {
        "entrypoint": "mock_entrypoint",
        "build_id": "mock_build_id",
        "compute_config_id": "mock_compute_config_id",
        "healthcheck_url": "mock_healthcheck_url",
        "access": access,
    }
    mock_logger = Mock()
    service_controller = ServiceController(log=mock_logger)
    if use_default_project:
        mock_infer_project_id = Mock(return_value="mock_default_project_id")
    else:
        mock_infer_project_id = Mock(return_value="mock_project_id")

    mock_validate_successful_build = Mock()

    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple(
        "anyscale.controllers.service_controller",
        infer_project_id=mock_infer_project_id,
    ), patch.object(
        ServiceController,
        "_get_maximum_uptime_minutes",
        return_value=maximum_uptime_minutes,
    ), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ), patch.multiple(
        "os.path", exists=Mock(return_value=True)
    ):
        service_controller.deploy(
            "mock_config_file", name="mock_name", description="mock_description",
        )

    service_controller.api_client.apply_service_api_v2_decorated_ha_jobs_apply_service_put.assert_called_once_with(
        CreateProductionService(
            name="mock_name",
            description="mock_description",
            project_id="mock_default_project_id"
            if use_default_project
            else "mock_project_id",
            config=ProductionJobConfig(  # noqa: PIE804
                **{
                    "entrypoint": "mock_entrypoint",
                    "build_id": "mock_build_id",
                    "compute_config_id": "mock_compute_config_id",
                }
            ),
            healthcheck_url="mock_healthcheck_url",
            access=access,
        )
    )
    assert mock_logger.info.call_count == 4


def test_generate_config_from_entrypoint(mock_auth_api_client):
    mock_logger = Mock()
    service_controller = ServiceController(log=mock_logger)
    entrypoint = ["python", "test.py"]
    name = "test_name"
    description = "test_description"

    test_compute_config_id = "mock_compute_config_id"
    test_build_id = "mock_build_id"

    mock_get_default_cluster_compute = Mock(
        return_value=Mock(id=test_compute_config_id)
    )
    mock_validate_successful_build = Mock()
    mock_get_default_cluster_env_build = Mock(return_value=Mock(id=test_build_id))
    with patch.multiple(
        "anyscale.controllers.job_controller",
        get_default_cluster_compute=mock_get_default_cluster_compute,
        validate_successful_build=mock_validate_successful_build,
        get_default_cluster_env_build=mock_get_default_cluster_env_build,
    ):
        job_config = service_controller.generate_config_from_entrypoint(
            entrypoint, name, description
        )

    assert job_config.entrypoint == "python test.py"
    assert job_config.name == name
    assert job_config.description == description


@pytest.mark.parametrize(
    ("ray_serve_config", "runtime_env"),
    [
        [{"runtime_env": {"pip": ["requests"], "working_dir": "."}}, None],
        [None, {"pip": ["requests"]}],
    ],
)
def test_generate_config_from_file(
    mock_auth_api_client, ray_serve_config: Optional[str], runtime_env: Optional[str]
):
    """
    This test checks that the config file for Service v1 and v2
    can be properly parsed.
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "entrypoint": "python test.py" if runtime_env else None,
        "healthcheck_url": "test.com" if runtime_env else None,
        "runtime_env": runtime_env,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }
    mock_validate_successful_build = Mock()
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        config = service_controller.generate_config_from_file(
            "test_ser8vice_config_file",
        )

        expected_service_config = ServiceConfig(
            name=name,
            description=description,
            ray_serve_config=ray_serve_config,
            entrypoint="python test.py" if runtime_env else None,
            runtime_env=runtime_env,
            healthcheck_url="test.com" if runtime_env else None,
            project_id=project_id,
            build_id=build_id,
            compute_config_id=compute_config_id,
            version="test_abc",
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )
    assert config == expected_service_config


@pytest.mark.parametrize("add_extra_field", [True, False])
def test_config_validation(add_extra_field, mock_auth_api_client):
    """
    This test checks that an error is thrown if extra/mispelled fields
    are added into the config file
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    extra_field_value = "*****"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": {},
        "entrypoint": "python test.py",
        "healthcheck_url": "test.com",
        "runtime_env": {},
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }

    if add_extra_field:
        extra_field_value = "*****"
        config_dict["extra_field"] = extra_field_value

    mock_validate_successful_build = Mock()
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        if add_extra_field:
            with pytest.raises(ValidationError):
                service_controller.generate_config_from_file(
                    "test_service_config_file", name=name, description=description,
                )
        else:
            service_controller.generate_config_from_file(
                "test_service_config_file", name=name, description=description,
            )


@pytest.mark.parametrize(
    ("user_provided_working_dir", "expected_working_dir_arg"),
    [
        (".", "."),
        ("./subdir/subdir2", "./subdir/subdir2"),
        ("/root_dir/subdir1", "/root_dir/subdir1"),
    ],
)
def test_generate_config_from_file_inside_workspace_working_dir_override(
    mock_auth_api_client,
    user_provided_working_dir: Optional[str],
    expected_working_dir_arg: str,
):
    """
    This test checks that if the service v2 is launched from the workspace without
    a working dir being specified, the method would back up the files into S3
    and update the working_dir of the service to point to the S3 location.
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    # Simulates that the user doesn't specify either working_dir or upload_path
    ray_serve_config = {
        "runtime_env": {"pip": ["requests"], "working_dir": user_provided_working_dir}
    }
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }
    mock_validate_successful_build = Mock()

    cloud_id = "test_cloud_id"
    org_id = "test_org_id"
    service_controller.api_client.get_decorated_cluster_api_v2_decorated_sessions_cluster_id_get.return_value = DecoratedsessionResponse(
        result=Mock(cloud=Mock(id=cloud_id))
    )

    s3_bucket_name = "test_s3_bucket_name"
    service_controller.api_client.get_cloud_with_cloud_resource_api_v2_clouds_with_cloud_resource_router_cloud_id_get.return_value = Mock(
        result=Mock(
            provider=CloudProviders.AWS, cloud_resource=Mock(aws_s3_id=s3_bucket_name),
        )
    )
    service_controller.api_client.get_user_info_api_v2_userinfo_get.return_value = Mock(
        result=Mock(organizations=[Mock(id=org_id)])
    )

    workspace_id = "Test_workspace_id"
    expected_working_dir = "s3://bucket"
    rewritten_runtime_env = {
        "pip": ["requests"],
        "working_dir": expected_working_dir,
    }
    upload_and_rewrite_working_dir_mock = Mock(return_value=rewritten_runtime_env)
    with patch.dict(
        os.environ,
        {
            "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": workspace_id,
            "ANYSCALE_SESSION_ID": "test_session_id",
        },
    ), patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple(
        "os.path", exists=Mock(return_value=True)
    ), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ), patch.multiple(
        "anyscale.utils.runtime_env",
        upload_and_rewrite_working_dir=upload_and_rewrite_working_dir_mock,
    ):
        config = service_controller.generate_config_from_file(
            "test_service_config_file",
        )

        expected_service_config = ServiceConfig(
            name=name,
            description=description,
            ray_serve_config={"runtime_env": rewritten_runtime_env},
            project_id=project_id,
            build_id=build_id,
            compute_config_id=compute_config_id,
            version="test_abc",
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )

    expected_s3_bucket_prefix = (
        f"s3://{s3_bucket_name}/{org_id}/{cloud_id}/workspace_snapshots/{workspace_id}"
    )

    call_args = upload_and_rewrite_working_dir_mock.call_args
    runtime_env_arg = call_args[0][0]
    assert runtime_env_arg["working_dir"] == expected_working_dir_arg
    assert runtime_env_arg["upload_path"].startswith(expected_s3_bucket_prefix)

    assert config == expected_service_config


@pytest.mark.parametrize(
    ("working_dir", "upload_path"),
    [
        ("https://github.com/anyscale/docs_examples/archive/refs/heads/main.zip", None),
        ("s3://already-a-uri/path", None),
    ],
)
def test_generate_config_from_file_inside_workspace_no_working_dir_override(
    mock_auth_api_client, working_dir: str, upload_path: Optional[str]
):
    """
    This test checks that if the service v2 is launched from the workspace with
    a remote working dir being specified, the method would inspect the remote url
    and not override it.
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    ray_serve_config = {
        "runtime_env": {
            "pip": ["requests"],
            "working_dir": working_dir,
            "upload_path": upload_path,
        }
    }
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"
    canary_percent = 100
    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": canary_percent,
        "rollout_strategy": rollout_strategy,
    }
    mock_validate_successful_build = Mock()

    workspace_id = "Test_workspace_id"
    with patch.dict(
        os.environ,
        {
            "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": workspace_id,
            "ANYSCALE_SESSION_ID": "test_session_id",
        },
    ), patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple(
        "os.path", exists=Mock(return_value=True)
    ), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        config = service_controller.generate_config_from_file(
            "test_service_config_file",
        )

        expected_service_config = ServiceConfig(
            name=name,
            description=description,
            ray_serve_config=ray_serve_config,
            project_id=project_id,
            build_id=build_id,
            compute_config_id=compute_config_id,
            version="test_abc",
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )

    assert config == expected_service_config


def test_service_submit_parse_logic(mock_auth_api_client) -> None:
    service_controller = ServiceController()
    service_controller.generate_config_from_entrypoint = Mock()  # type: ignore
    service_controller.generate_config_from_file = Mock()  # type: ignore
    service_controller.deploy_from_config = Mock()  # type: ignore

    # We are not in a workspace, so entrypoint should not be allowed
    with pytest.raises(click.ClickException):
        service_controller.deploy(
            "file", None, None, entrypoint=["entrypoint"], is_entrypoint_cmd=False
        )

    with pytest.raises(click.ClickException):
        service_controller.deploy(
            "file", None, None, entrypoint=["entrypoint"], is_entrypoint_cmd=True
        )

    with pytest.raises(click.ClickException):
        service_controller.deploy(
            "file",
            None,
            None,
            entrypoint=["entrypoint", "commands"],
            is_entrypoint_cmd=True,
        )

    # Simulate a workspace
    with patch.dict(
        os.environ, {"ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": "fake_workspace_id"}
    ), patch.multiple(
        "anyscale.controllers.service_controller",
        override_runtime_env_for_local_working_dir=Mock(
            return_value={"rewritten": True}
        ),
    ):
        # Fails due to is_entrypoint_cmd being False
        with pytest.raises(click.ClickException):
            service_controller.deploy(
                "file",
                name=None,
                description=None,
                entrypoint=["entrypoint"],
                is_entrypoint_cmd=False,
            )

        mock_config = Mock()
        service_controller.generate_config_from_file.return_value = mock_config
        service_controller.deploy(
            "file", name=None, description=None, entrypoint=[], is_entrypoint_cmd=False
        )
        service_controller.generate_config_from_file.assert_called_once_with(
            "file", name=None, description=None, healthcheck_url=None
        )
        service_controller.deploy_from_config.assert_called_once_with(mock_config)
        service_controller.generate_config_from_file.reset_mock()
        service_controller.deploy_from_config.reset_mock()

        mock_config = Mock()
        service_controller.generate_config_from_entrypoint.return_value = mock_config
        service_controller.deploy(
            "file",
            name=None,
            description=None,
            entrypoint=["entrypoint"],
            is_entrypoint_cmd=True,
        )
        service_controller.generate_config_from_entrypoint.assert_called_once_with(
            ["file", "entrypoint"], name=None, description=None, healthcheck_url=None
        )
        service_controller.deploy_from_config.assert_called_once_with(mock_config)


@pytest.mark.parametrize("override_name", [None, "override_service_name"])
@pytest.mark.parametrize("override_version", [None, "override_version"])
@pytest.mark.parametrize("override_canary_percent", [None, 0, 40])
@pytest.mark.parametrize("override_rollout_strategy", [None, "IN_PLACE"])
def test_service_rollout(
    mock_auth_api_client,
    override_name: Optional[str],
    override_version: Optional[str],
    override_canary_percent: Optional[int],
    override_rollout_strategy: Optional[str],
) -> None:
    """
    This tests that when we submit a Service config with ray_serve_config,
    we submit both the Service v1 and Service v2 configs.
    Please reference services_internal_api_models.py for more details.
    """
    service_controller = ServiceController()

    name = "test_service_name"
    description = "mock_description"
    ray_serve_config = {"runtime_env": {"pip": ["requests"], "working_dir": "."}}
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    rollout_strategy = "ROLLOUT"

    config_dict = {
        "name": name,
        "description": description,
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": "test_abc",
        "canary_percent": 100,
        "rollout_strategy": rollout_strategy,
    }
    mock_validate_successful_build = Mock()
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.controllers.job_controller",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_controller.rollout(
            "test_ser8vice_config_file",
            name=override_name,
            version=override_version,
            canary_percent=override_canary_percent,
            rollout_strategy=override_rollout_strategy,
        )

    service_v2_config = ApplyProductionServiceV2APIModel(
        name=override_name if override_name else name,
        description=description,
        build_id=build_id,
        compute_config_id=compute_config_id,
        project_id=project_id,
        ray_serve_config=ray_serve_config,
        version=override_version if override_version else "test_abc",
        canary_percent=override_canary_percent
        if override_canary_percent is not None
        else 100,
        rollout_strategy=override_rollout_strategy
        if override_rollout_strategy
        else "ROLLOUT",
    )

    service_controller.api_client.apply_service_v2_api_v2_services_v2_apply_put.assert_called_once_with(
        service_v2_config
    )


class TestGetServiceID:
    def test_invalid_inputs(self, mock_auth_api_client):
        service_controller = ServiceController()

        with pytest.raises(
            click.ClickException,
            match="Service ID, name, or config file must be specified.",
        ):
            service_controller.get_service_id()

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id", service_name="test-name"
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id",
                service_config_file="test-config-file-path",
                project_id=None,
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_id="test-id",
                service_name="test-name",
                service_config_file="test-config-file-path",
            )

        with pytest.raises(click.ClickException, match="Only one of"):
            service_controller.get_service_id(
                service_name="test-name", service_config_file="test-config-file-path"
            )

        with pytest.raises(click.ClickException, match="not found"):
            service_controller.get_service_id(
                service_config_file="/fake/service/config/path"
            )

    def test_service_id_provided(self, mock_auth_api_client):
        """If the ID is provided, the same ID should always be returned and no API calls made."""
        service_controller = ServiceController()

        assert service_controller.get_service_id(service_id="test-id") == "test-id"
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_not_called()

    def test_name_provided_no_matching_service(self, mock_auth_api_client):
        """If a name is provided but no services match it, an exception should be raised."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[], metadata=Mock(next_paging_token=None),
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"No service with name '{TEST_SERVICE_NAME}' was found",
        ):
            service_controller.get_service_id(service_config_file="test-config-file")

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )

    def test_name_provided_single_matching_service(self, mock_auth_api_client):
        """If a name is provided and one service matches it, its ID should be returned."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        assert (
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test the config file codepath.
        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )
        assert (
            service_controller.get_service_id(service_config_file="test-config-file")
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )

    def test_name_provided_multiple_matching_services(self, mock_auth_api_client):
        """If a name is provided and multiple services match it, an exception should be raised."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id="test-service-id-1"), Mock(id="test-service-id-2")],
            metadata=Mock(next_paging_token=None),
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(service_name=TEST_SERVICE_NAME)

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        mock_service_config = Mock()
        mock_service_config.configure_mock(name=TEST_SERVICE_NAME, project_id=None)
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        with pytest.raises(
            click.ClickException,
            match=f"There are multiple services with name '{TEST_SERVICE_NAME}'.",
        ):
            service_controller.get_service_id(service_config_file="test-config-file")

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
        )

    def test_no_project_id(self, mock_auth_api_client):
        """If there is no project_id, the results should not be filtered based on project."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(service_name=TEST_SERVICE_NAME)
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                count=10, creator_id=None, name=TEST_SERVICE_NAME, project_id=None,
            )

    def test_project_id_in_service_config(self, mock_auth_api_client):
        """If the project_id is passed in the service config it should be used as a filter."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        mock_service_config = Mock()
        mock_service_config.configure_mock(
            name=TEST_SERVICE_NAME, project_id="test-project-id"
        )
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(
                    service_config_file="test-config-file"
                )
                == TEST_SERVICE_ID
            )
            service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
                count=10,
                creator_id=None,
                name=TEST_SERVICE_NAME,
                project_id="test-project-id",
            )

    def test_override_project_id(self, mock_auth_api_client):
        """If the project_id is passed directly, the passed value should be used to filter."""
        service_controller = ServiceController()

        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[Mock(id=TEST_SERVICE_ID)], metadata=Mock(next_paging_token=None),
        )

        # Test overriding when getting service ID by name.
        assert (
            service_controller.get_service_id(
                service_name=TEST_SERVICE_NAME, project_id="overridden-project-id"
            )
            == TEST_SERVICE_ID
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id="overridden-project-id",
        )
        service_controller.api_client.list_services_api_v2_services_v2_get.reset_mock()

        # Test overriding when getting service ID by config file.
        mock_service_config = Mock()
        mock_service_config.configure_mock(
            name=TEST_SERVICE_NAME, project_id="test-project-id"
        )
        service_controller._read_service_config_file = Mock(
            return_value=mock_service_config,
        )

        mock_infer_project_id = Mock(return_value="mock_default_project_id")
        with patch.multiple(
            "anyscale.controllers.service_controller",
            infer_project_id=mock_infer_project_id,
        ):
            assert (
                service_controller.get_service_id(
                    service_config_file="test-config-file",
                    project_id="overridden-project-id",
                )
                == TEST_SERVICE_ID
            )

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            count=10,
            creator_id=None,
            name=TEST_SERVICE_NAME,
            project_id="overridden-project-id",
        )


def test_terminate_service(mock_auth_api_client):
    """Test that v2 IDs go through v2 API, v1 through v1."""
    service_controller = ServiceController()

    service_controller.terminate("service2_abc123")
    service_controller.api_client.terminate_service_api_v2_services_v2_service_id_terminate_post.assert_called_once_with(
        "service2_abc123",
    )

    service_controller.job_controller._resolve_job_object = Mock(
        return_value=Mock(id="service_abc123"),
    )
    service_controller.terminate("service_abc123")
    service_controller.api_client.terminate_job_api_v2_decorated_ha_jobs_production_job_id_terminate_post.assert_called_once_with(
        "service_abc123",
    )


class TestListServices:
    def test_list_service_v1_with_id(self, mock_auth_api_client):
        service_controller = ServiceController()
        service_controller.list(service_id="service_v1_id")

        service_controller.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get.assert_called_once_with(
            "service_v1_id",
        )

    def test_list_service_v2_with_id(self, mock_auth_api_client):
        service_controller = ServiceController()
        service_controller.list(service_id="service2_id")

        service_controller.api_client.get_service_api_v2_services_v2_service_id_get.assert_called_once_with(
            "service2_id",
        )

    def test_list_service_unexpected_error(self, mock_auth_api_client):
        service_controller = ServiceController()
        service_controller.list(service_id="service2_id")

        service_controller.api_client.list_services_api_v2_services_v2_get.side_effect = Exception(
            "Request failed! Unexpected exception!"
        )

        service_controller.list()

        service_controller.job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.assert_not_called()

    def test_list_service_error_fallback(self, mock_auth_api_client):
        """
        We fallback to list Services v1 if the execption indicates that the
        FF is turned off for the user.
        """
        service_controller = ServiceController()
        service_controller.list(service_id="service2_id")

        service_controller.api_client.list_services_api_v2_services_v2_get.side_effect = Exception(
            "Services v2 is not enabled for your organization."
        )
        service_controller.job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[], metadata=Mock(next_paging_token=None),
        )

        service_controller.list()

        service_controller.job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.assert_called_once_with(
            creator_id=None,
            name=None,
            project_id=None,
            type_filter="SERVICE",
            archive_status="NOT_ARCHIVED",
            count=10,
        )

    @pytest.mark.parametrize("name", [None, "service_name"])
    @pytest.mark.parametrize("project_id", [None, "test_project_id"])
    @pytest.mark.parametrize("created_by_me", [True, False])
    def test_list_service(
        self,
        mock_auth_api_client,
        name: Optional[str],
        project_id: Optional[str],
        created_by_me: bool,
    ):
        service_controller = ServiceController()
        service_controller.api_client.list_services_api_v2_services_v2_get.return_value = DecoratedlistserviceapimodelListResponse(
            results=[], metadata=Mock(next_paging_token=None),
        )
        if created_by_me:
            creator_id = "test_user_id"
            service_controller.api_client.get_user_info_api_v2_userinfo_get.return_value = Mock(
                result=Mock(id=creator_id)
            )
        else:
            creator_id = None

        service_controller.list(
            name=name, project_id=project_id, created_by_me=created_by_me
        )

        service_controller.api_client.list_services_api_v2_services_v2_get.assert_called_once_with(
            creator_id=creator_id, name=name, project_id=project_id, count=10,
        )

    def test_list_service_with_pagination(
        self, mock_auth_api_client,
    ):
        service_controller = ServiceController()

        next_paging_token = "test"
        list_return_values = [
            DecoratedlistserviceapimodelListResponse(
                results=[],
                metadata=ListResponseMetadata(
                    total=11, next_paging_token=next_paging_token
                ),
            ),
            DecoratedlistserviceapimodelListResponse(
                results=[],
                metadata=ListResponseMetadata(total=11, next_paging_token=None),
            ),
        ]
        service_controller.api_client.list_services_api_v2_services_v2_get.side_effect = (
            list_return_values
        )

        service_controller.list()

        calls = [
            call(creator_id=None, name=None, project_id=None, count=10),
            call(
                creator_id=None,
                name=None,
                project_id=None,
                count=10,
                paging_token=next_paging_token,
            ),
        ]
        service_controller.api_client.list_services_api_v2_services_v2_get.assert_has_calls(
            calls
        )


def test_rollback(mock_auth_api_client):
    """Test that v2 IDs go through v2 API, v1 is rejected."""
    service_controller = ServiceController()

    service_controller.rollback("service2_abc123")
    service_controller.api_client.rollback_service_api_v2_services_v2_service_id_rollback_post.assert_called_once_with(
        "service2_abc123",
    )

    with pytest.raises(
        click.ClickException, match="rollback is only supported for v2 services"
    ):
        service_controller.rollback("service_abc123")


def test_archive(mock_auth_api_client):
    """Test that v1 IDs go through v1 API, v2 is rejected."""
    service_controller = ServiceController()

    with pytest.raises(
        click.ClickException, match="archive is not currently supported for v2 services"
    ):
        service_controller.archive("service2_abc123")

    service_controller.job_controller._resolve_job_object = Mock(
        return_value=Mock(id="service_abc123"),
    )
    service_controller.archive("service_abc123")
    service_controller.api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post.assert_called_once_with(
        "service_abc123",
    )
